/*
	Copyright 2018-2019 Giovanni Franza - HB9EIK 

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <dlfcn.h>
#include <syslog.h>
#include <sys/time.h>
#include <dirent.h> 
#include <sys/signal.h> 

#include "struct.h"
#include "TMATE2.h"

#define XtNumber(x) sizeof(x)/sizeof(x[0])

static work works;
static setup setupd;
static plotter plotterd;
static color colors;
static cat cats;
static int id;

static void (*auxdown[12])( void *, int );
static void (*auxwrite[12])( void *, int );
static void (*down[12])( void *, int );
static void (*up[12])( void *, int );
static int kbbutton[12];
static int *value[12];
static void *kbwork=NULL;
static int tasti[12];
static int oldtasti[12];

void (* ReadConfig)( workp works );
void (* WriteInit)( workp works );
int pickObj( int x, int y );

void onexit( void ) {
	kill( works.socatPid, SIGKILL );
	WriteInit( &works );
}

void auxDown( void *works, int value ) {
	((workp)works)->buttons[value].activated*=-1;
}

void kbCallback( int j ) {
	if( auxdown[j] ) {
		auxdown[j]( kbwork, kbbutton[j] );
	}
	if( auxwrite[j] ) {
		auxwrite[j]( kbwork, *value[j] );
	}
	if( down[j] ) {
		down[j]( kbwork, kbbutton[j]>=0?kbbutton[j]:1 );
	}
	if( up[j] ) {
		up[j]( kbwork, kbbutton[j]>=0?kbbutton[j]:-1 );
	}
}
void kbCB( void *works, int j ) {
	kbCallback( j );
}
int kbAddCallback( int keyNumber, int Button, int *Value, void *Work, void (*AuxDown)( void *, int ), void (*Auxwrite)( void *, int ), void (*Down)( void *, int ), void (*Up)( void *, int ) ) {
        if( keyNumber <0 && keyNumber>=12 ) {
                syslog( LOG_ERR, "invalid keyNumber in kbAddCallback: %d not in 0-11", keyNumber );
                return -1;
        }
        up[keyNumber]=Up;
        down[keyNumber]=Down;
        auxdown[keyNumber]=AuxDown;
        auxwrite[keyNumber]=Auxwrite;
        kbbutton[keyNumber]=Button;
        value[keyNumber]=Value;
        kbwork=Work;
        return 0;
}

void kbSet( workp works ) {
	int j, k;
	for( j=0; j<12; j++ ) {
		for( k=0; k<works->tmateNbuttons; k++ ) {
			if( works->setup->kbkeys[j] == works->tmateButtons[k] ) {
				works->setup->kbkeysidx[j]=k;
				break;
			}
		}
		works->kbaddcallback( (void *)works, j );
	}
}

void kbaddcallback( void *works, int num ) {
	int k;
	k=((workp)works)->setup->kbkeys[num]-1;
	if( !(((workp)works)->buttons+k)->isEncoder ) {
		if( (((workp)works)->buttons+k)->down == NULL ) {
			if( (((workp)works)->buttons+k)->up == NULL ) {
				if( (((workp)works)->buttons+k)->tmate==1 ) {
					((workp)works)->kbAddCallback( num, 1, NULL, (void*)works, NULL, NULL, (((workp)works)->buttons+k)->write, NULL );
				} else if( (((workp)works)->buttons+k)->tmate==2 ) {
					((workp)works)->kbAddCallback( num, 1, NULL, (void*)works, NULL, NULL, (((workp)works)->buttons+k)->dxdn, NULL );
				} else {
					((workp)works)->kbAddCallback( num, (((workp)works)->buttons+k)->tmate-2, NULL, (void*)works, NULL, NULL, (((workp)works)->buttons+k)->write, NULL );
				}
			} else {
				if( (((workp)works)->buttons+k)->tmate==1 ) {
					((workp)works)->kbAddCallback( num, -1, NULL, (void*)works, NULL, NULL, (((workp)works)->buttons+k)->write, (((workp)works)->buttons+k)->write );
				} else {
					((workp)works)->kbAddCallback( num, -1, NULL, (void*)works, NULL, NULL, (((workp)works)->buttons+k)->write, NULL );
				}
			}
		} else {
			((workp)works)->kbAddCallback( num, k, &((((workp)works)->buttons+k)->activated), (void*)works, ((workp)works)->auxdown, (((workp)works)->buttons+k)->write, (((workp)works)->buttons+k)->down, (((workp)works)->buttons+k)->up );
		}
	}
}

void display(void) {
	int j, k, obj;
	GLint wp[4];
	static int x=-1;
	static int y=-1;
	static struct timeval Drawstart, Drawstop;
	// vedo cosa colorare
	if( x!=works.plotter->mousePosX || y!=works.plotter->mousePosY ) {
		x=works.plotter->mousePosX;
		y=works.plotter->mousePosY;
		glPushMatrix();
		obj = pickObj( works.plotter->mousePosX, works.plotter->mousePosY );
		if( (obj==-2 || obj==-3) && works.plotter->newFreq>0 ) {
			works.plotter->newFreq=works.plotter->mousePosX*800/works.realxSize;
			works.plotter->lastY=obj;
		}
		if( obj > 0 && obj < 1000 ) {
			(works.buttons[obj]).over=1;
		}
		for( j=0; j<works.nbuttons; j++ ) {
			if( j!=obj ) {
				(works.buttons[j]).over=0;
			}
		}
		glPopMatrix();
	}

	// pulisco il buffer
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

	// decido da dove guardare
	gluLookAt( 0, 0, 6000, 0, 0, 0, 0, 1, 0 );

	// disegno tutto l'ambaradan
	works.DrawData( &works );

	// forzo la visualizzazione
	glutSwapBuffers();
	glFlush();
	works.xPos=glutGet((GLenum)GLUT_WINDOW_X)-works.xDelta;
	works.yPos=glutGet((GLenum)GLUT_WINDOW_Y)-works.yDelta;
	if( works.xPos<0 ) {
		works.xPos=0;
	}
	if( works.yPos<0 ) {
		works.yPos=0;
	}
        glGetIntegerv (GL_VIEWPORT, wp );
	works.xSize=wp[2];
	works.ySize=wp[3];
	glutPostRedisplay();
	if( works.plotter->layer != 16 ) {
		gettimeofday( &Drawstop, NULL );
		works.Drawrythm=1000.0*(Drawstop.tv_sec-Drawstart.tv_sec)+0.001*(Drawstop.tv_usec-Drawstart.tv_usec);
		gettimeofday( &Drawstart, NULL );
	}
}

// ridisegno in caso di ridimensionamento
void reshape( int w, int h ) {
	static int isFirst=1;

	// orrible hack to cope with glut insane metrics
	if( isFirst ) {
		works.xDelta=glutGet((GLenum)GLUT_WINDOW_X)-glutGet((GLenum)GLUT_INIT_WINDOW_X);
		works.yDelta=glutGet((GLenum)GLUT_WINDOW_Y)-glutGet((GLenum)GLUT_INIT_WINDOW_Y);
		if( works.xDelta<0 ) {
			works.xDelta=0;
		}
		if( works.yDelta<0 ) {
			works.yDelta=0;
		}
		isFirst=0;
	}
	works.realxSize=w;
	works.realySize=h;
	glViewport ( 0,  0, (GLsizei)w, (GLsizei)h );
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glOrtho ( (GLdouble)000, (GLdouble)800, (GLdouble)00, (GLdouble)600, -4000, 15000 );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
}

int pickObj( int x, int y ) {

	int top, left, height, width;
	int h,w;
	int j;

	w=x*800/works.realxSize;
	h=600-y*600/works.realySize;

	for( j=0; j<works.nbuttons; j++ ) {
		if( ((works.buttons[j].layer) & (works.plotter->layer))==0 ) {
			continue;
		}
		if(  works.buttons[j].left <= w && works.buttons[j].left+works.buttons[j].width >= w && works.buttons[j].top >= h && works.buttons[j].top-works.buttons[j].height <= h ) {
			if(  works.buttons[j].top-works.buttons[j].height*0.5 >= h ) {
				works.lowerClick=1;
			} else {
				works.lowerClick=0;
			}
			return j;
		}
	}

	left=works.plotter->WfallXpos;
	width=works.plotter->WfallXsize;
	top=works.plotter->WfallYpos;
	height=works.plotter->WfallYsize;
	if(  left <= w && left+width >= w && top >= h && top-height <= h ) {
		return -2;
	}

	left=works.plotter->FFTXpos;
	width=works.plotter->FFTXsize;
	top=works.plotter->FFTYpos;
	height=works.plotter->FFTYsize;
	if(  left <= w && left+width >= w && top >= h && top-height <= h ) {
		return -3;
	}

	left=works.plotter->FreqXpos;
	width=works.plotter->FreqXsize;
	top=works.plotter->FreqYpos;
	height=works.plotter->FreqYsize;
	if(  left <= w && left+width >= w && top >= h && top-height <= h ) {
		return -4;
	}

	left=works.plotter->MODXpos;
	width=works.plotter->MODXsize;
	top=works.plotter->MODYpos;
	height=works.plotter->MODYsize;
	if(  left <= w && left+width >= w && top >= h && top-height <= h ) {
		return -5;
	}

	left=works.plotter->RSSIXpos;
	width=works.plotter->RSSIXsize;
	top=works.plotter->RSSIYpos;
	height=works.plotter->RSSIYsize;
	if(  left <= w && left+width >= w && top >= h && top-height <= h ) {
		return -6;
	}

	left=works.plotter->SyntXpos;
	width=works.plotter->SyntXsize;
	top=works.plotter->SyntYpos;
	height=works.plotter->SyntYsize;
	if(  left <= w && left+width >= w && top >= h && top-height <= h ) {
		return -7;
	}

	left=works.plotter->OTHXpos;
	width=works.plotter->OTHXsize;
	top=works.plotter->OTHYpos;
	height=works.plotter->OTHYsize;
	if(  left <= w && left+width >= w && top >= h && top-height <= h ) {
		return -8;
	}

	return -1;
} 

// gestione tastiera
void myKeyb( unsigned char key, int x, int y ) {
		
	if( key == works.setup->End ) {
		works.unterminated=0;
		sleep( 1 );
		exit( 0 );
	}

	if( key=='R' || key=='r' ) {
		works.modeChange( &works, 6 );
	}
	if( key=='C' || key=='c' ) {
		works.modeChange( &works, 4 );
	}
	if( key=='L' || key=='l' ) {
		works.modeChange( &works, 2 );
	}
	if( key=='A' || key=='a' ) {
		works.modeChange( &works, 1 );
	}
	if( key=='U' || key=='u' ) {
		works.modeChange( &works, 3 );
	}
	if( key=='F' || key=='f' ) {
		works.modeChange( &works, 5 );
	}

	if( key==13 && strlen( works.plotter->data) ) {
		works.AddCipher( &works, 'e' );
	}
	if( key==8 ) {
		works.AddCipher( &works, '<' );
	}
	if( key=='#' ) {
		works.AddCipher( &works, '#' );
	}
	if( key=='+'  ) {
		works.AddCipher( &works, 0 );
		works.AddCipher( &works, 0 );
		works.AddCipher( &works, 0 );
	}
	if( key=='k' && strlen( works.plotter->data)  ) {
		works.AddCipher( &works, 'k' );
	}
	if( key>='0' && key<='9' ) {
		works.AddCipher( &works, key-'0' );
	}
}

// Gestione tasti speciali
static struct timeval oldtime;
static struct timeval ora;
void mySpecial( int key, int x, int y ) {
	int freq, inc;
	static int oldkey=0;
	double uno, due;
	static int isFirst=1;

	if( isFirst ) {
		isFirst=0;
		oldtime.tv_sec=0;
		oldtime.tv_usec=0;
	}
	if( key==oldkey ) {
		gettimeofday( &ora, NULL );
		uno=1000.0*ora.tv_sec+0.001*ora.tv_usec;
		due=1000.0*oldtime.tv_sec+0.001*oldtime.tv_usec;
		if( uno < due+50.0 ) {
			return;
		}
	}
	gettimeofday( &oldtime, NULL );
	oldkey=key;

	switch( key ) {
		case GLUT_KEY_F1:
			kbCallback( 0 );
			break;
		case GLUT_KEY_F2:
			kbCallback( 1 );
			break;
		case GLUT_KEY_F3:
			kbCallback( 2 );
			break;
		case GLUT_KEY_F4:
			kbCallback( 3 );
			break;
		case GLUT_KEY_F5:
			kbCallback( 4 );
			break;
		case GLUT_KEY_F6:
			kbCallback( 5 );
			break;
		case GLUT_KEY_F7:
			kbCallback( 6 );
			break;
		case GLUT_KEY_F8:
			kbCallback( 7 );
			break;
		case GLUT_KEY_F9:
			kbCallback( 8 );
			break;
		case GLUT_KEY_F10:
			kbCallback( 9 );
			break;
		case GLUT_KEY_F11:
			kbCallback( 10 );
			break;
		case GLUT_KEY_F12:
			kbCallback( 11 );
			break;
		case GLUT_KEY_UP:	
			if( works.freqIncr == 0 ) {
				works.freqIncr = 1;
			} else {
				works.freqIncr *= 10;
				if( works.freqIncr >= 10000000 ) {
					works.freqIncr = 10000000;
				}
			}
			break;
		case GLUT_KEY_DOWN:
			if( works.freqIncr == 0 ) {
				works.freqIncr = 1;
			} else {
				works.freqIncr /= 10;
				if( works.freqIncr <= 1 ) {
					works.freqIncr = 1;
				}
			}
			break;
		case GLUT_KEY_RIGHT:
			inc=works.freqIncr;
			freq=works.serParms->freq+inc;
			freq=inc*(freq/inc);
			works.sintonia( &works, freq );
			break;
		case GLUT_KEY_LEFT:
			inc=works.freqIncr;
			freq=works.serParms->freq-inc;
			freq=inc*ceil((double)freq/(double)inc);
			works.sintonia( &works, freq );
			break;
		default:
			return;
	}
	pickObj( x, y );
}

void mouseEnter( int enter ) {
	int j;
	if( enter==GLUT_LEFT ) {
		for( j=0; j<works.nbuttons; j++ ) {
			(works.buttons[j]).over=0;
		}
	}
}

void mouseMove( int x, int y ) {
	works.plotter->mousePosX=x;
	works.plotter->mousePosY=y;
}
#define GLUT_WHEEL_UP 3
#define GLUT_WHEEL_DOWN 4
void mouseWheel( int button, int state, int x, int y ) {
	int *spi;
	int obj;
	int mod;
	int inc;
        int j;
        int group;
        int toggle;
	int freq;
	int rc;

	if( state == GLUT_UP ) {
 		obj = pickObj( x, y );
		if( obj==-8 && button == GLUT_LEFT) {
			(works.setAB)(&works,0);
			return;
		}
		if( obj==-2 && button == GLUT_LEFT && works.plotter->newFreq>0) {
			freq=works.serParms->freq-80000+(works.plotter->newFreq-works.plotter->FFTXpos)*160000/(works.plotter->FFTXsize-1);
			inc=works.setup->AutoSnap;
			freq=(double)inc*round((double)freq/(double)inc);
			works.sintonia( &works, freq );
			works.plotter->newFreq=-1;
		}
		if( obj==-2 && button == GLUT_RIGHT_BUTTON && works.plotter->newFreq>0) {
			freq=works.serParms->freq-80000+(works.plotter->newFreq-works.plotter->FFTXpos)*160000/(works.plotter->FFTXsize-1);
			inc=works.setup->AutoSnap;
			freq=(double)inc*round((double)freq/(double)inc);
			works.sintonia( &works, freq+works.serParms->ritFreq );
			works.plotter->newFreq=-1;
		}
		if( obj!=-2 ) {
			works.plotter->newFreq=-1;
		}
		if( obj < 0 || obj > 1000 ) {
			return;
		}
		if((works.buttons[obj]).isEncoder) {
			return;
		}
		(works.buttons[obj]).over=1;
		if( button == GLUT_LEFT || ( button == GLUT_RIGHT_BUTTON && (works.buttons[obj]).dxdn) ) {
			if((works.buttons[obj]).up) {
				(works.buttons[obj].up)(&works,works.buttons[obj].activated);
			}
		}
	}
	else if( state == GLUT_DOWN ) {
 		obj = pickObj( x, y );
		if( obj==-2 || obj==-3 ) {
			if( button == GLUT_LEFT || button == GLUT_RIGHT_BUTTON ) {
				works.plotter->newFreq=x*800/works.realxSize;
				works.plotter->lastY=obj;
			} else  {
				mod = glutGetModifiers();
				inc = works.freqIncr;
				switch ( mod ) {
					case GLUT_ACTIVE_SHIFT:
						inc *= 100 ;
						break;
					case GLUT_ACTIVE_CTRL:
						inc *= 10 ;
						break;
					case GLUT_ACTIVE_ALT:
						inc /= 10 ;
						break;
					default: 
						inc = works.freqIncr;
						break;
				}
				if( button == GLUT_WHEEL_DOWN ) {
					freq=works.serParms->freq-inc;
					freq=inc*ceil((double)freq/(double)inc);
					works.sintonia( &works, freq );
				} else if( button == GLUT_WHEEL_UP ) {
					freq=works.serParms->freq+inc;
					freq=inc*(freq/inc);
					works.sintonia( &works, freq );
				}
			}
		}
		if( obj==-6 ) {
			works.irttmed=0;
			works.irttmax=0;
		}
		if( obj < 0 || obj > 1000 ) {
			return;
		}
		if((works.buttons[obj]).isEncoder) {
			if( works.setup->isTouch ) {
				if( button == GLUT_LEFT_BUTTON ) {
					obj = pickObj( x, y );
					if( obj < 0 || obj > 1000 ) {
						return;
					}
					if( works.lowerClick ) {
						(works.buttons[obj].wheelDn)(&works,works.buttons[obj].activated);
					} else {
						(works.buttons[obj].wheelUp)(&works,works.buttons[obj].activated);
					}
				} else if( button == GLUT_RIGHT_BUTTON ) {
					obj = pickObj( x, y );
					if( obj < 0 || obj > 1000 ) {
						return;
					}
					if( works.lowerClick ) {
						(works.buttons[obj].down)(&works,works.buttons[obj].activated);
					} else {
						(works.buttons[obj].up)(&works,works.buttons[obj].activated);
					}
				}
			} else {
				if( button == GLUT_WHEEL_UP ) {
					obj = pickObj( x, y );
					if( obj < 0 || obj > 1000 ) {
						return;
					}
					if((works.buttons[obj]).wheelUp) {
						(works.buttons[obj].wheelUp)(&works,works.buttons[obj].activated);
					}
				} else if( button == GLUT_WHEEL_DOWN ) {
					obj = pickObj( x, y );
					if( obj < 0 || obj > 1000 ) {
						return;
					}
					if((works.buttons[obj]).wheelDn) {
						(works.buttons[obj].wheelDn)(&works,works.buttons[obj].activated);
					}
				} else if( button == GLUT_LEFT_BUTTON ) {
					if( (works.buttons[obj]).down ) {
						(works.buttons[obj]).down( (void *)&works, works.buttons[obj].activated );
					}
				} else if( button == GLUT_RIGHT_BUTTON ) {
					if((works.buttons[obj]).up) {
						(works.buttons[obj].up)(&works,works.buttons[obj].activated);
					}
				}
			}
		} else {
			(works.buttons[obj]).over=2;
			group=works.buttons[obj].group;
			toggle=works.buttons[obj].toggle;
			works.buttons[obj].pressed=1;
			works.buttons[obj].modified=1;
			if( (works.buttons[obj]).read ) {
				(works.buttons[obj]).read( (void *)&works, obj );
			}
			if( toggle==1 || toggle==2 ) {
				if( button == GLUT_LEFT ) {
					works.buttons[obj].activated*=-1;
					for( j=0; j<works.nbuttons; j++ ) {
						if( j!=obj ) {
							works.buttons[j].pressed=0;
							if( works.buttons[j].group == group && j!=obj ) {
								if( works.buttons[j].activated>0 ) {
									works.buttons[j].modified=1;
									works.buttons[j].activated*=-1;
								}
							}
						}
					}
				}
			} else if( toggle==4 ) {
				if( button == GLUT_LEFT ) {
					if( (works.buttons[obj]).down ) {
						(works.buttons[obj]).down( (void *)&works, works.buttons[obj].activated );
					}
					if( (works.buttons[obj]).write ) {
						(works.buttons[obj]).write( (void *)&works, works.buttons[obj].activated );
					}
				}
			} else {
				if( button == GLUT_LEFT ) {
					if( works.buttons[obj].activated<0 ) {
						works.buttons[obj].activated*=-1;
					}
					for( j=0; j<works.nbuttons; j++ ) {
						if( j!=obj ) {
							works.buttons[j].pressed=0;
							if( works.buttons[j].group == group && j!=obj ) {
								if( works.buttons[j].activated>0 ) {
									works.buttons[j].modified=1;
									works.buttons[j].activated*=-1;
								}
							}
						}
					}
				}
			}
			if( toggle!=4 ) {
				if( button == GLUT_LEFT ) {
					if( (works.buttons[obj]).write ) {
						(works.buttons[obj]).write( (void *)&works, works.buttons[obj].activated );
					}
					if( (works.buttons[obj]).down ) {
						(works.buttons[obj]).down( (void *)&works, works.buttons[obj].activated );
					}
				}
			}
			if( button == GLUT_RIGHT_BUTTON ) {
				if((works.buttons[obj]).dxdn) {
					(works.buttons[obj].dxdn)(&works,works.buttons[obj].activated);
				}
			}
			if( button == GLUT_MIDDLE_BUTTON ) {
				if((works.buttons[obj]).centerdn) {
					(works.buttons[obj].centerdn)(&works,works.buttons[obj].activated);
				}
			}
		}
	}
}

int main(int argc, char** argv) {

        pthread_t pthread2, pthread3, pthread4, pthread5, pthread6, pthread7, pthread8, pthread9;
        pthread_attr_t attr2, attr3, attr4, attr5, attr6, attr7, attr8, attr9;
	int thrLeggi, thrScrivi;
	int err;
	float Ydata[4*1024];
	pthread_mutex_t wmut=PTHREAD_MUTEX_INITIALIZER;
	serialParms serParms;
	linearParms linParms;
	myStrings mStrings;
	int j,k,l;
	void *libHandle;
	char *error;
	struct dirent **namelist;
	char qq[300];
	struct sched_param prio2;
	struct sched_param prio3;
	struct sched_param prio4;
	struct sched_param prio5;
	struct sched_param prio6;
	struct sched_param prio7;
	struct sched_param prio8;
	struct sched_param prio9;
	int  (* SetTimerResolution)( long, unsigned char, long* );
	long CurrentResolution;
	pthread_mutex_t audioFftMut=PTHREAD_MUTEX_INITIALIZER;;
	pthread_cond_t audioFftCond=PTHREAD_COND_INITIALIZER;;
        FILE *fp;
        char buf[100];
        char *pp;

	memset( &works, 0, sizeof( works ) );

	works.nActions=0;
	works.nDrawings=0;
	works.nLanguages=0;
	setupd.langue=0;

	works.argv0 = argv[0];

	signal(SIGIO, SIG_IGN);
	signal(SIGPIPE, SIG_IGN);

	j = scandir(".", &namelist, NULL, alphasort);
	if (j < 0) {
		syslog( LOG_ERR, "no files in directory" );
		exit(1);
	} else {
		while (j--) {
			sprintf( qq, "./%s", namelist[j]->d_name );
			if( !strncmp( namelist[j]->d_name, "libActions", 10 ) ) {
				if( works.nActions<3 ) {
					dlerror();
					works.libActions[works.nActions] = dlopen( qq, RTLD_LAZY);
					if (works.libActions[works.nActions]) {
						works.nActions++;
						syslog( LOG_ERR, "opened library libActions: %s", qq);
					} else {
						syslog( LOG_ERR, "error opening library libActions: %s %s", qq, dlerror());
					}
				}
			} else if( !strncmp( namelist[j]->d_name, "libConfig", 9 ) ) {
				dlerror();
				libHandle = dlopen( qq,  RTLD_LAZY );
				if (!libHandle) {
					syslog( LOG_ERR, "error opening library libConfig: %s %s", qq, dlerror());
					exit(1);
				} else {
					syslog( LOG_ERR, "opened library libConfig: %s", qq);
				}
				dlerror();
				ReadConfig = dlsym( libHandle, "ReadConfig" );
				if ((error = dlerror()) != NULL)  {
					syslog( LOG_ERR, "error retrieving function ReadConfig: %s %s", qq, error);
					exit(1);
				}
				dlerror();
				WriteInit = dlsym( libHandle, "WriteInit" );
				if ((error = dlerror()) != NULL)  {
					syslog( LOG_ERR, "error retrieving function WriteInit: %s", error);
					exit(1);
				}
			} else if( !strncmp( namelist[j]->d_name, "libDrawings", 11 ) ) {
				if( works.nDrawings<3 ) {
					dlerror();
					works.libDrawings[works.nDrawings] = dlopen( qq, RTLD_LAZY);
					if (works.libDrawings[works.nDrawings]) {
						works.nDrawings++;
					} else {
						syslog( LOG_ERR, "error opening library libDrawing: %s %s", qq, dlerror());
					}
				}
			} else if( !strncmp( namelist[j]->d_name, "langue-", 7 ) ) {
				for( k=7; namelist[j]->d_name[k]; k++ ) {
					if( !strcmp( namelist[j]->d_name+k, ".txt" ) ) {
						namelist[j]->d_name[k]='\0';
						break;
					}
				}
				if( works.nLanguages<20 ) {
					strcpy( works.Languages[works.nLanguages], namelist[j]->d_name+7 );
					works.nLanguages++;
				}
			} 
			free(namelist[j]);
		}
		free(namelist);
	}
	if (!works.nActions) {
		syslog( LOG_ERR, "error opening library libActions: no libraries found" );
		exit(1);
	}
	if (!works.nDrawings) {
		syslog( LOG_ERR, "error opening library libDrawing: no libraries found" );
		exit(1);
	}
	libHandle = dlopen( "ntdll.dll",  RTLD_LAZY );
	dlerror();
	SetTimerResolution = dlsym( libHandle, "NtSetTimerResolution" );
	if( !dlerror() )  {
		SetTimerResolution( 5000, 1, &CurrentResolution );
	}
	works.plotter=&plotterd;
        for( j=0; j<XtNumber(Ydata); j++ ) {
                Ydata[j]=0.0;
        }
	works.strings=&mStrings;
	locale( &works, NULL );
	works.colors=&colors;
	works.cat=&cats;
        works.plotter->Ydata=Ydata;
	works.wmut=&wmut;
	pthread_mutex_init( works.wmut, NULL );
	works.started=0;
	works.serReceived=0;
	works.serParms=&serParms;
	works.linParms=&linParms;
	linParms.stato=-1;
	works.freqRitIncr=10;
	works.freqIncr=100;
	works.auxVolIncr=10;
	works.saveVol=-1;
	works.setup=&setupd;
	works.auxdown=auxDown;
	works.tmateaddcallback=tmateaddcallback;
	works.kbaddcallback=kbaddcallback;
	works.kbAddCallback=kbAddCallback;
	works.kbcb=kbCB;
	(*ReadConfig)( &works );
	locale( &works, works.setup->myLangue );
	if( works.setup->enableTmate == '1' ) {
		controltmate( &works );
	}
	kbSet( &works );
	works.unterminated=1;
	works.thrWsock=-1;
	works.audioErrors=0;
	works.numSameError=0;
	works.freqChanged.tv_sec=0;
	works.freqChanged.tv_usec=0;
	works.ritChanged.tv_sec=0;
	works.ritChanged.tv_usec=0;
        works.volumeChanged.tv_sec=0;
        works.volumeChanged.tv_usec=0;
        works.auxVolChanged.tv_sec=0;
        works.auxVolChanged.tv_usec=0;
        works.modeChanged.tv_sec=0;
        works.modeChanged.tv_usec=0;
        works.bwChanged.tv_sec=0;
        works.bwChanged.tv_usec=0;
        works.thresholdChanged.tv_sec=0;
        works.thresholdChanged.tv_usec=0;
        works.agcMlvlChanged.tv_sec=0;
        works.agcMlvlChanged.tv_usec=0;
        works.agcAlvlChanged.tv_sec=0;
        works.agcAlvlChanged.tv_usec=0;
        works.nrChanged.tv_sec=0;
        works.nrChanged.tv_usec=0;
        works.nbChanged.tv_sec=0;
        works.nbChanged.tv_usec=0;
        works.anChanged.tv_sec=0;
        works.anChanged.tv_usec=0;
        works.squelchChanged.tv_sec=0;
        works.squelchChanged.tv_usec=0;
        works.lastFFTshown.tv_sec=0;
        works.lastFFTshown.tv_usec=0;
        works.lastFFTreceived.tv_sec=0;
        works.lastFFTreceived.tv_usec=0;
	works.audioFftMut=&audioFftMut;
	pthread_mutex_init( works.audioFftMut, NULL );
	works.audioFftCond=&audioFftCond;
	works.noAudio=0;
	works.setup->newDev=1;

	if( argc>=2 && strlen( argv[1] ) ) {
		strcpy( works.setup->server, argv[1] );
		strcpy( works.server, works.setup->server );
	}
	if( argc>=3 && strlen( argv[2] ) ) {
		strcpy( works.passphrase, argv[2] );
	}
	if( argc>=4 && strlen( argv[3] ) ) {
		works.port = atoi( argv[3] );
	}

	works.plotter->layer=works.setup->startLayer;
	works.plotter->baseLayer=works.setup->startLayer;
	readPng( &works, "smeter", works.smeter, &works.height_smeter, &works.width_smeter );
	readPng( &works, "wsmeter", works.wsmeter, &works.height_wsmeter, &works.width_wsmeter );
	readPng( &works, "meter", works.meter, &works.height_meter, &works.width_meter );
	readPng( &works, "wmeter", works.wmeter, &works.height_wmeter, &works.width_wmeter );
	readPng( &works, "button", works.buttonTexture, &works.height_button, &works.width_button );
	readPng( &works, "encoder1", works.encoder1Texture, &works.height_encoder1, &works.width_encoder1 );
	readPng( &works, "encoder2", works.encoder2Texture, &works.height_encoder2, &works.width_encoder2 );
	readPng( &works, "encoder3", works.encoder3Texture, &works.height_encoder3, &works.width_encoder3 );
	readPng( &works, "encoder4", works.encoder4Texture, &works.height_encoder4, &works.width_encoder4 );

	glutInit( &argc, argv );
	//Request double buffer display mode.
	glutInitDisplayMode( GLUT_ALPHA | GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH );
	works.realxSize=works.xSize;
	works.realySize=works.ySize;
	glutInitWindowSize( works.xSize, works.ySize ); 
	glutInitWindowPosition( works.xPos, works.yPos );
	id=glutCreateWindow( "EladRemote V1.77 - 2020-07-21- Developed by HB9EIK for ELAD" );
	glLoadIdentity();
	glClearColor (1.0, 1.0, 1.0, 0.0);
	glutDisplayFunc( display ); 
	glutReshapeFunc( reshape );
	glutKeyboardFunc( myKeyb );
	glutSpecialFunc( mySpecial );
	glutMouseFunc( mouseWheel );
	glutPassiveMotionFunc( mouseMove );
	glutEntryFunc( mouseEnter );
	glutMotionFunc( mouseMove );
	glEnable( GL_NORMALIZE );
	glEnable( GL_DEPTH_TEST );
	glEnable( GL_LIGHTING );
	glEnable( GL_POINT_SMOOTH );
	glEnable( GL_SMOOTH );
	atexit( onexit );

	fp = popen ( "socat -d -d pty,raw,echo=0 pty,raw,echo=0 >/dev/null 2>/dev/null &\necho $!", "r" );
	pp = fgets( buf, sizeof( buf ), fp );
	works.socatPid = atoi( pp );
	fclose( fp );
 
	if( setupd.enableWrite == '1' ) {
		works.decoder = opus_decoder_create( works.rate, 2, &err );
		if( err<0 ) {
			syslog( LOG_ERR, "failed to create decoder: %s", opus_strerror( err ) );
			works.rxcompress=0;
		}
	}
	if( setupd.enableRead == '1' ) {
		works.encoder = opus_encoder_create( works.rate, 2, OPUS_APPLICATION_AUDIO, &err );
		if( err<0 ) {
			syslog( LOG_ERR, "failed to create an encoder: %s", opus_strerror( err ) );
			works.txcompress=0;
		}
		err = opus_encoder_ctl( works.encoder, OPUS_SET_BITRATE( works.txcompress ) );
		if( err<0 && works.txcompress ) {
			syslog( LOG_ERR, "failed to set bitrate: %s", opus_strerror( err ) );
			works.txcompress=0;
		}
	}
	if( setupd.enableWrite == '1' || setupd.enableRead == '1' ) {
		pthread_attr_init( &attr2 );
		thrLeggi=pthread_create( &pthread2, &attr2, th_audio, (void *)&works );
		prio2.sched_priority=80;
		err=pthread_setschedparam( pthread2, SCHED_RR, &prio2 );
		if( err ) {
			syslog( LOG_ERR, "pthread_setschedparam pthread 2 failed Error %d errno %d", err, errno );
		}
	}
	for( ;works.started<2; ) {
		if( !works.unterminated ) {
			return 0;
		}
		if( setupd.enableWrite != '1' || setupd.enableRead != '1' ) {
			break;
		}
		usleep( 10000 );
		glutMainLoopEvent();
	}

	usleep( 10000 );
	if( setupd.webSocket == '2' ) {
		pthread_attr_init( &attr4 );
		pthread_create( &pthread4, &attr4, Wclient, (void *)&works );
		prio4.sched_priority=80;
		err=pthread_setschedparam( pthread4, SCHED_RR, &prio4 );
		if( err ) {
			syslog( LOG_ERR, "pthread_setschedparam pthread 4 failed Error %d errno %d", err, errno );
		}
	}
	if( setupd.enableTime == '1' ) {
		pthread_attr_init( &attr5 );
		pthread_create( &pthread5, &attr5, timer, (void *)&works );
		prio5.sched_priority=20;
		err=pthread_setschedparam( pthread5, SCHED_RR, &prio5 );
		if( err ) {
			syslog( LOG_ERR, "pthread_setschedparam pthread 5 failed Error %d errno %d", err, errno );
		}
	}
	if( setupd.enableCW == '1' ) {
		pthread_attr_init( &attr6 );
		pthread_create( &pthread6, &attr6, keyer, (void *)&works );
		prio6.sched_priority=99;
		err=pthread_setschedparam( pthread6, SCHED_RR, &prio6 );
		if( err ) {
			syslog( LOG_ERR, "pthread_setschedparam pthread 6 failed Error %d errno %d", err, errno );
		}
		pthread_attr_init( &attr7 );
		pthread_create( &pthread7, &attr7, oscillator, (void *)&works );
		prio7.sched_priority=90;
		err=pthread_setschedparam( pthread7, SCHED_RR, &prio7 );
		if( err ) {
			syslog( LOG_ERR, "pthread_setschedparam pthread 7 failed Error %d errno %d", err, errno );
		}
	}
	if( setupd.enableAudioFFT == '1' ) {
		pthread_attr_init( &attr8 );
		pthread_create( &pthread8, &attr8, audioFFT, (void *)&works );
		prio8.sched_priority=80;
		err=pthread_setschedparam( pthread8, SCHED_RR, &prio8 );
		if( err ) {
			syslog( LOG_ERR, "pthread_setschedparam pthread 8 failed Error %d errno %d", err, errno );
		}
	}
	if( setupd.enableFIFO == '1' ) {
		pthread_attr_init( &attr9 );
		pthread_create( &pthread9, &attr9, fifo, (void *)&works );
		prio9.sched_priority=20;
		err=pthread_setschedparam( pthread9, SCHED_RR, &prio9 );
		if( err ) {
			syslog( LOG_ERR, "pthread_setschedparam pthread 9 failed Error %d errno %d", err, errno );
		}
	}

	for( ;works.unterminated; ) {
		glutMainLoopEvent();
	}
	opus_encoder_destroy( works.encoder );
	opus_decoder_destroy( works.decoder );
	return 0;
}
