/*
	Copyright 2018-2019 Giovanni Franza - HB9EIK 

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <wchar.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <libusb.h>
#include "pthread.h"
#include "TMATE2.h"

static const int VENDOR_ID = 0x1721;
static const int PRODUCT_ID = 0x0614;
struct libusb_device_handle *handle = NULL;
static unsigned char buf_t[255];
static unsigned char buf_a[255];
static pthread_t pthread1, pthread2;
static pthread_attr_t attr1, attr2;
static int thrRead, thrWrite;
static int unterminated=1;
static libusb_context *usb_context = NULL;

static void (*auxdown[12])( void *, int );
static void (*auxwrite[12])( void *, int );
static void (*down[12])( void *, int );
static void (*up[12])( void *, int );
static int button[12];
static int *value[12];
static void *work=NULL;
static int tasti[12];
static int oldtasti[12];
static int enabled[12];

unsigned char NumericSegmentMatrix[91]={
//    a				  b              c            d           e           f          g
      0, 			  0, 			     0, 		      0, 	      0, 		   0, 		  0, 
 nCOM0|SEG4  ,   nCOM1|SEG4,	nCOM2|SEG4,  nCOM3|SEG3, nCOM2|SEG3, nCOM1|SEG3, nCOM0|SEG3,  // digit 1
 nCOM0|SEG6  ,   nCOM1|SEG6,	nCOM2|SEG6,  nCOM3|SEG5, nCOM2|SEG5, nCOM1|SEG5, nCOM0|SEG5,  // digit 2
 nCOM0|SEG8  ,   nCOM1|SEG8,	nCOM2|SEG8,  nCOM3|SEG7, nCOM2|SEG7, nCOM1|SEG7, nCOM0|SEG7,  // digit 3
 nCOM0|SEG10 ,   nCOM1|SEG10,	nCOM2|SEG10, nCOM3|SEG9, nCOM2|SEG9, nCOM1|SEG9, nCOM0|SEG9,  // digit 4
 nCOM0|SEG12 ,   nCOM1|SEG12,	nCOM2|SEG12, nCOM3|SEG11,nCOM2|SEG11,nCOM1|SEG11,nCOM0|SEG11, // digit 5
 nCOM0|SEG14 ,   nCOM1|SEG14,	nCOM2|SEG14, nCOM3|SEG13,nCOM2|SEG13,nCOM1|SEG13,nCOM0|SEG13, // digit 6
 nCOM0|SEG16 ,   nCOM1|SEG16,	nCOM2|SEG16, nCOM3|SEG15,nCOM2|SEG15,nCOM1|SEG15,nCOM0|SEG15, // digit 7
 nCOM0|SEG18 ,   nCOM1|SEG18,	nCOM2|SEG18, nCOM3|SEG17,nCOM2|SEG17,nCOM1|SEG17,nCOM0|SEG17, // digit 8
 nCOM0|SEG20 ,   nCOM1|SEG20,	nCOM2|SEG20, nCOM3|SEG19,nCOM2|SEG19,nCOM1|SEG19,nCOM0|SEG19, // digit 9
 nCOM7|SEG27 ,   nCOM6|SEG27,	nCOM5|SEG27, nCOM4|SEG27,nCOM5|SEG28,nCOM6|SEG28,nCOM7|SEG28, // digit 10 // digit 17
 nCOM7|SEG25 ,   nCOM6|SEG25,	nCOM5|SEG25, nCOM4|SEG25,nCOM5|SEG26,nCOM6|SEG26,nCOM7|SEG26, // digit 11 // digit 18
 nCOM7|SEG23 ,   nCOM6|SEG23,	nCOM5|SEG23, nCOM4|SEG23,nCOM5|SEG24,nCOM6|SEG24,nCOM7|SEG24, // digit 12 // digit 19
};

short LCDNumberMask[]={
     SEGA|SEGB|SEGC|SEGD|SEGE|SEGF,        			// 0
          SEGB|SEGC,                       			// 1
     SEGA|SEGB    |SEGD|SEGE      |SEGG,   			// 2
     SEGA|SEGB|SEGC|SEGD          |SEGG,   			// 3
     SEGB|SEGC               |SEGF|SEGG,   			// 4
     SEGA     |SEGC|SEGD     |SEGF|SEGG,   			// 5
     SEGA     |SEGC|SEGD|SEGE|SEGF|SEGG,  		 	// 6
     SEGA|SEGB|SEGC,                       			// 7
     SEGA|SEGB|SEGC|SEGD|SEGE|SEGF|SEGG,   			// 8
     SEGA|SEGB|SEGC|SEGD     |SEGF|SEGG,   			// 9
	 SEGA|SEGB|SEGC|	 SEGE|SEGF|SEGG,   			// A 10
	 SEGA|SEGB|			 SEGE|SEGF|SEGG,   			// P
	 SEGA|SEGB|SEGC|	 SEGE|SEGF,              	// M
	 SEGA			|SEGD|SEGE|SEGF,        		// C 13
	 	      SEGB|SEGC|	      SEGE|SEGF|SEGG,   // H
                   SEGC                        ,    // i
              SEGB|                        SEGF,    // "
	      SEGG,   						 			// - 17
		SEGC|SEGD|SEGE|SEGF|SEGG,   			    // b 18
		SEGC|SEGD|SEGE,			 		            // u 19 
			 SEGE|SEGF				,		        // l 20
     0 ,                                     		// digit spento
			 SEGD|SEGE|SEGF|SEGG,		  			// t 22
	 SEGA|        SEGE|SEGF|SEGG,		  			// F 23
	 SEGA|SEGC|SEGD|   SEGF|SEGG,		  			// S 24
	           SEGD|SEGE|SEGF,		      			// L 25
     SEGA     |SEGD|SEGE|SEGF|SEGG,  	  			// E 26
     SEGE|SEGG,   						 			// r 27
     SEGE|SEGF|SEGG,   						 			// K 28
};

static void read_callback( struct libusb_transfer *transfer ) {
	int rc;
	int j;
	short enc;

	if (transfer->status == LIBUSB_TRANSFER_COMPLETED) {
		memcpy( buf_a, transfer->buffer, transfer->actual_length );
		memcpy( (char *)&enc, buf_a+5, 2 );
		tasti[9]=enc;
		memcpy( (char *)&enc, buf_a+3, 2 );
		tasti[10]=enc;
		memcpy( (char *)&enc, buf_a+1, 2 );
		tasti[11]=(-enc);
		tasti[0]=1-(buf_a[7]&0x01);	
		tasti[1]=1-((buf_a[7]&0x02)>>1);	
		tasti[2]=1-((buf_a[7]&0x04)>>2);	
		tasti[3]=1-((buf_a[7]&0x08)>>3);	
		tasti[4]=1-((buf_a[7]&0x10)>>4);	
		tasti[5]=1-((buf_a[7]&0x20)>>5);	
		tasti[6]=1-((buf_a[7]&0x40)>>6);	
		tasti[7]=1-((buf_a[7]&0x80)>>7);	
		tasti[8]=1-(buf_a[8]&0x01);	
		for( j=0; j<6; j++ ) {
			if( enabled[j] && tasti[j]!=oldtasti[j] ) {
				oldtasti[j]=tasti[j];
				if( tasti[j] ) {
					if( auxdown[j] ) {
						auxdown[j]( work, button[j] );
					}
					if( auxwrite[j] ) {
						auxwrite[j]( work, *value[j] );
					}
					if( down[j] ) {
						down[j]( work, button[j]>=0?button[j]:1 );
					}
				} else {
					if( up[j] ) {
						up[j]( work, button[j]>=0?button[j]:-1 );
					}
				}
			}
		}
		for( j=6; j<9; j++ ) {
			if( enabled[j] && tasti[j]!=oldtasti[j] ) {
				oldtasti[j]=tasti[j];
				if( tasti[j] ) {
					// 14-j means that F7 is E1 clickedd, F8 is E2 clickked, F9 is Tune cliccked
					if( auxdown[14-j] ) {
						auxdown[14-j]( work, button[14-j] );
					}
					if( auxwrite[14-j] ) {
						auxwrite[14-j]( work, *value[14-j] );
					}
					if( down[14-j] ) {
						down[14-j]( work, button[14-j]>=0?button[14-j]:1 );
					}
				} else {
					if( up[14-j] ) {
						up[14-j]( work, button[14-j]>=0?button[14-j]:-1 );
					}
				}
			}
		}
		for( j=9; j<11; j++ ) {
			if( enabled[j] && tasti[j]!=oldtasti[j] ) {
				rc=0;
				if( tasti[j]>oldtasti[j]+1 ) {
					rc=1;
				}
				if( tasti[j]<oldtasti[j]-1 ) {
					rc=2;
				}
				if( rc==1 ) {
					if( up[j] ) {
						up[j]( work, button[j] );
					}
					oldtasti[j]=tasti[j];
				} 
				if( rc==2 ) {
					if( down[j] ) {
						down[j]( work, button[j] );
					}
					oldtasti[j]=tasti[j];
				}
			}
		}
		if( enabled[j] && tasti[11]!=oldtasti[11] ) {
			rc=0;
			if( tasti[11]>oldtasti[11] ) {
				rc=1;
			}
			if( tasti[11]==-(1<<15) && oldtasti[11]==(1<<15)-1 ) {
				rc=1;
			}
			if( tasti[11]==(1<<15)-1 && oldtasti[11]==-(1<<15) ) {
				rc=0;
			}
			if( rc ) {
				if( up[11] ) {
					up[11]( work, button[11] );
				}
			} else {
				if( down[11] ) {
					down[11]( work, button[11] );
				}
			}
			oldtasti[11]=tasti[11];
		}
	} else if( transfer->status == LIBUSB_TRANSFER_CANCELLED || transfer->status == LIBUSB_TRANSFER_NO_DEVICE ) {
		//unterminated = 0;
		return;
	}
	rc = libusb_submit_transfer( transfer );
	if (rc != 0) {
		syslog( LOG_ERR, "Unable to submit URB. libusb error code: %d", rc );
		//unterminated = 0;
	}
	return;
}

void *th_write( void *a ) {
	int res;
	memset( buf_t, 0, sizeof( buf_t ) );
	for( ; unterminated; ) {
		buf_t[0] = 0x00;
		int actual_length;
                res = libusb_interrupt_transfer( handle,
                        0x01,
                        (unsigned char*)buf_t+1,
                        64,
                        &actual_length, 1000);
		if( res < 0 ) {
			syslog( LOG_ERR, "unable to write to device 1721:0614" );
		}
		usleep( 10000 );
	}
	return a;
}
void *th_read( void *a ) {
        static struct libusb_transfer *transfer;
	static unsigned char buf[255];
	int length=64;

	usleep( 100000 );
	transfer = libusb_alloc_transfer( 0 );
	libusb_fill_bulk_transfer( transfer, handle, 0x81, buf, length, (libusb_transfer_cb_fn)read_callback, NULL, 5000 );

	libusb_submit_transfer( transfer );

	while( unterminated ) {
		int res;
		res = libusb_handle_events( usb_context );
		if (res < 0) {
			syslog( LOG_ERR, "read_thread(): libusb reports error # %d\n", res);
			if (res != LIBUSB_ERROR_BUSY && res != LIBUSB_ERROR_TIMEOUT && res != LIBUSB_ERROR_OVERFLOW && res != LIBUSB_ERROR_INTERRUPTED) {
				// unterminated = 0;
				break;
			}
		}
	}
	return a;
}
int TMate2OpenConnection( void ) {
	int rc;
	int j;

	syslog( LOG_INFO, "libTMATE2 v0.10 2019-06-04 by HB9EIK for ELAD starting" );

	for( j=0; j<12; j++ ) {
		up[j]=NULL;
		auxdown[j]=NULL;
		auxwrite[j]=NULL;
		down[j]=NULL;
		value[j]=NULL;
		button[j]=0;
		oldtasti[j]=0;
		enabled[j]=0;
	}

	rc = libusb_init( &usb_context );
	if( rc < 0 ) {
		syslog( LOG_ERR, "unable to initialize lib_usb-1.0" );
		return -1;
	}
	syslog( LOG_INFO, "lib_usb-1.0 initialized" );

	handle = libusb_open_device_with_vid_pid(NULL, VENDOR_ID, PRODUCT_ID );
	if (handle == NULL) {
		syslog( LOG_ERR, "unable to open device %x:%x", VENDOR_ID, PRODUCT_ID );
		return -1;
	}
	syslog( LOG_INFO, "TMATE device %x:%x opened", VENDOR_ID, PRODUCT_ID );

	libusb_detach_kernel_driver(handle, 0 );
	syslog ( LOG_INFO, "kernel detached\n" );

	rc = libusb_claim_interface( handle, 0 );
	if( rc < 0 ) {
		syslog( LOG_ERR, "libusb_claim_interface error %d", rc );
	}
	syslog( LOG_INFO, "libusb claimed interface" );
	unterminated = 1;
	pthread_attr_init( &attr1 );
	thrWrite=pthread_create( &pthread1, &attr1, th_write, NULL );
	pthread_attr_init( &attr2 );
	thrRead=pthread_create( &pthread2, &attr2, th_read, NULL );
	//sleep( 1 );
	return 0; 
}
int TMate2CloseConnection( void ) {
	int j;
	for( j=0; j<12; j++ ) {
		up[j]=NULL;
		down[j]=NULL;
		button[j]=0;
		oldtasti[j]=0;
	}

	unterminated = 0;
	libusb_exit( usb_context );
	syslog( LOG_INFO, "libTMATE2 ended" );
	return 0;
}
int TMate2ReadValue( int *enc1, int *enc2, int *enc3, int *tasti ) {
	short enc;
	memcpy( (char *)&enc, buf_a+5, 2 );
	*enc1=enc;
	memcpy( (char *)&enc, buf_a+3, 2 );
	*enc2=enc;
	memcpy( (char *)&enc, buf_a+1, 2 );
	*enc3=enc;
	tasti[0]=1-(buf_a[7]&0x01);	
	tasti[1]=1-((buf_a[7]&0x02)>>1);	
	tasti[2]=1-((buf_a[7]&0x04)>>2);	
	tasti[3]=1-((buf_a[7]&0x08)>>3);	
	tasti[4]=1-((buf_a[7]&0x10)>>4);	
	tasti[5]=1-((buf_a[7]&0x20)>>5);	
	tasti[6]=1-((buf_a[7]&0x40)>>6);	
	tasti[7]=1-((buf_a[7]&0x80)>>7);	
	tasti[8]=1-(buf_a[8]&0x01);	
	return 0;
}
int TMate2ReadVersion(char *H1,char *H2,char *F1,char *F2) {
	unsigned char buf[255];
	int res=0;
	int len;
        len = libusb_get_string_descriptor(handle, 3, 0x0409, buf, 255);
	if( len > 22 ) {
		*H1=buf[6];
		*H2=buf[10];
		*F1=buf[18];
		*F2=buf[22];
	} else {
		res=-1;
	}
	return res;
}
void TMate2WriteNumberOnMainDisplay(unsigned char *LCDVector, int Value) {
	int j;
	int segment;
	int common;
	int digit = 1;
	int number = 8;
	char stringa[10];
	unsigned char lcd_mask;
	if( Value<0 ) {
		Value*= -1;
	}
	sprintf( stringa, "%9d", Value );
	for( digit=9; digit>=1; digit-- ) {
		if( stringa[digit-1]==' ' ) {
			number=BLANK;
		} else {
			number=stringa[digit-1]-'0';
		}
		short mask = 1;
		for( j = 0; j < 7; j++) {
			lcd_mask = NumericSegmentMatrix[digit*7+j];
			segment = lcd_mask & 0x1f;
			common = lcd_mask >> 5;
			if ((mask & LCDNumberMask[number]) != 0) {
				LCDVector[segment] |= (1 << common);
			} else {
				LCDVector[segment] &= ~(1 << common);
			}
			mask = (short)(mask << 1);
		}
	}
}
void TMate2WriteNumberOnSmeterDisplay(unsigned char *LCDVector, int SmeterValue) {
	int j;
	int segment;
	int common;
	int digit = 1;
	int number = 8;
	char stringa[4];
	unsigned char lcd_mask;
	if( SmeterValue<0 ) {
		SmeterValue*= -1;
	}
	sprintf( stringa, "%3d", SmeterValue%1000 );
	for( digit=12; digit>=10; digit-- ) {
		if( stringa[digit-10]==' ' ) {
			number=BLANK;
		} else {
			number=stringa[digit-10]-'0';
		}
		short mask = 1;
		for( j = 0; j < 7; j++) {
			lcd_mask = NumericSegmentMatrix[digit*7+j];
			segment = lcd_mask & 0x1f;
			common = lcd_mask >> 5;
			if ((mask & LCDNumberMask[number]) != 0) {
				LCDVector[segment] |= (1 << common);
			} else {
				LCDVector[segment] &= ~(1 << common);
			}
			mask = (short)(mask << 1);
		}
	}
}
void TMate2SetSegment(unsigned char *LCDVector, int lcd_mask, unsigned char status) {
	int segment;
	int common;
        segment = lcd_mask & 0x1f;
        common = lcd_mask >> 5;
        if ((status) != 0) {
                LCDVector[segment] |= (unsigned char)(1 << common);
        } else {
                LCDVector[segment] &= (unsigned char)~(1 << common);
        }
}
void TMate2SetLcd( unsigned char *bufout) {
	memset( buf_t, 0, sizeof( buf_t ) );
	memcpy( buf_t+1, bufout, sizeof( buf_t )-1 ) ;
}

void TMate2SetStatus( t2statusp st ) {
	static unsigned char buf[255];
	memset( buf, 0, sizeof( buf ) );
	if( st->lp ) {
		TMate2SetSegment( buf, SEG_LP, 1 );
	}
	if( st->att ) {
		TMate2SetSegment( buf, SEG_ATT, 1 );
		TMate2SetSegment( buf, SEG_ATT_1, 1 );
	}
	if( st->tx ) {
		TMate2SetSegment( buf, SEG_TX, 1 );
		if( st->power >=1000 ) {
			TMate2SetSegment( buf, SEG_mW_W, 1 );
			TMate2WriteNumberOnSmeterDisplay( buf, st->power/1000 );
		} else {
			TMate2SetSegment( buf, SEG_mW_m, 1 );
			TMate2SetSegment( buf, SEG_mW_W, 1 );
			TMate2WriteNumberOnSmeterDisplay( buf, st->power );
		}
	}
	if( st->rx ) {
		TMate2SetSegment( buf, SEG_RX, 1 );
		TMate2SetSegment( buf, SEG_DBM, 1 );
			if( st->smeter<0 ) {
				TMate2SetSegment( buf, SEG_DB_MINUS, 1 );
				TMate2WriteNumberOnSmeterDisplay( buf, -st->smeter );
				if( st->smeter > -121 ) {
					TMate2SetSegment( buf, SMETER_SEG1, 1 );
					TMate2SetSegment( buf, SEG_1, 1 );
				}
				if( st->smeter > -115 ) {
					TMate2SetSegment( buf, SMETER_SEG2, 1 );
				}
				if( st->smeter > -109 ) {
					TMate2SetSegment( buf, SMETER_SEG3, 1 );
					TMate2SetSegment( buf, SEG_3, 1 );
				}
				if( st->smeter > -103 ) {
					TMate2SetSegment( buf, SMETER_SEG4, 1 );
				}
				if( st->smeter > - 97 ) {
					TMate2SetSegment( buf, SMETER_SEG5, 1 );
					TMate2SetSegment( buf, SEG_5, 1 );
				}
				if( st->smeter > - 91 ) {
					TMate2SetSegment( buf, SMETER_SEG6, 1 );
				}
				if( st->smeter > - 85 ) {
					TMate2SetSegment( buf, SMETER_SEG7, 1 );
					TMate2SetSegment( buf, SEG_7, 1 );
				}
				if( st->smeter > - 79 ) {
					TMate2SetSegment( buf, SMETER_SEG8, 1 );
				}
				if( st->smeter > - 73 ) {
					TMate2SetSegment( buf, SMETER_SEG9, 1 );
					TMate2SetSegment( buf, SEG_9, 1 );
				}
				if( st->smeter > - 63 ) {
					TMate2SetSegment( buf, SMETER_SEG10, 1 );
				}
				if( st->smeter > - 53 ) {
					TMate2SetSegment( buf, SMETER_SEG11, 1 );
					TMate2SetSegment( buf, SEG_20, 1 );
				}
				if( st->smeter > - 43 ) {
					TMate2SetSegment( buf, SMETER_SEG12, 1 );
				}
				if( st->smeter > - 33 ) {
					TMate2SetSegment( buf, SMETER_SEG13, 1 );
					TMate2SetSegment( buf, SEG_40, 1 );
				}
				if( st->smeter > - 23 ) {
					TMate2SetSegment( buf, SMETER_SEG14, 1 );
				}
				if( st->smeter > - 13 ) {
					TMate2SetSegment( buf, SMETER_SEG15, 1 );
					TMate2SetSegment( buf, SEG_60, 1 );
				}
			} else {
				TMate2WriteNumberOnSmeterDisplay( buf, st->smeter );
			}
                        
	}
	if( st->vfo ) {
		TMate2SetSegment( buf, SEG_VFO, 1 );
		TMate2SetSegment( buf, SEG_A, 1 );
	} else {
		TMate2SetSegment( buf, SEG_VFO, 1 );
		TMate2SetSegment( buf, SEG_B, 1 );
	}
	if( st->nr ) {
		TMate2SetSegment( buf, SEG_NR, 1 );
	}
	if( st->nb ) {
		TMate2SetSegment( buf, SEG_NB, 1 );
	}
	if( st->an ) {
		TMate2SetSegment( buf, SEG_AN, 1 );
	}
	TMate2WriteNumberOnMainDisplay( buf, st->frequency );
	TMate2SetSegment( buf, SEG_K, 0 );
	TMate2SetSegment( buf, SEG_HZ, 1 );
	if( st->am ) {
		TMate2SetSegment( buf, SEG_AM, 1 );
	}
	if( st->lsb ) {
		TMate2SetSegment( buf, SEG_LSB, 1 );
	}
	if( st->usb ) {
		TMate2SetSegment( buf, SEG_USB, 1 );
	}
	if( st->cwplus ) {
		TMate2SetSegment( buf, SEG_CW, 1 );
		TMate2SetSegment( buf, SEG_CW_PLUS, 1 );
	}
	if( st->cwminus ) {
		TMate2SetSegment( buf, SEG_CW, 1 );
		TMate2SetSegment( buf, SEG_CW_MINUS, 1 );
	}
	if( st->fm ) {
		TMate2SetSegment( buf, SEG_FM, 1 );
	}
	TMate2SetSegment( buf, SEG_E1, 1 );
	if( st->e1vol ) {
		TMate2SetSegment( buf, SEG_VOL, 1 );
	}
	if( st->e1sql ) {
		TMate2SetSegment( buf, SEG_SQL, 1 );
	}
	if( st->e1nr ) {
		TMate2SetSegment( buf, SEG_T33, 1 );
	}
	if( st->e1nb ) {
		TMate2SetSegment( buf, SEG_T35, 1 );
	}
	if( st->e1an ) {
		TMate2SetSegment( buf, SEG_T36, 1 );
	}
	TMate2SetSegment( buf, SEG_E2, 1 );
	if( st->e2rit ) {
		TMate2SetSegment( buf, SEG_RIT, 1 );
	}
	buf[0x26] |= 0x01; // attiva manopola vfo rovescia
	buf[0x20] |= 0x01; // Led accensione
	//buf[0x20] |= 0x02; // Led rosso
	if( st->tx ) {
		memcpy( buf+0x21, st->txcolor, 3 );
	}
	if( st->rx ) {
		memcpy( buf+0x21, st->rxcolor, 3 );
	}
	TMate2SetLcd( buf );
}
int TMate2AddCallback( int keyNumber, int Button, int *Value, void *Work, void (*AuxDown)( void *, int ), void (*Auxwrite)( void *, int ), void (*Down)( void *, int ), void (*Up)( void *, int ) ) {
	if( keyNumber <0 && keyNumber>=12 ) {
		syslog( LOG_ERR, "invalid keyNumber in TMate2AddCallback: %d not in 0-11", keyNumber );
		return -1;
	}	
	enabled[keyNumber]=0;
	usleep( 5000 );
	up[keyNumber]=Up;
	down[keyNumber]=Down;
	auxdown[keyNumber]=AuxDown;
	auxwrite[keyNumber]=Auxwrite;
	button[keyNumber]=Button;
	value[keyNumber]=Value;
	work=Work;
	enabled[keyNumber]=1;
	usleep( 5000 );
	return 0;
}

