/*
	Copyright 2018-2019 Giovanni Franza - HB9EIK 

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <syslog.h>
#include <string.h>
#include <math.h>
#define XtNumber(x) sizeof(x)/sizeof(x[0])

#include "struct.h"

void compute( char *in, int num, int pitch ) {
	int j;
	short sample;
	double i, q, t, g;
	double iu, qu, tu, gu;
	double id, qd, td, gd;
	g=pitch*2*M_PI/48000;
	gd=(pitch-4)*2*M_PI/48000;
	gu=(pitch+4)*2*M_PI/48000;
	i=iu=id=q=qu=qd=0.0;
	for( j=0; j< num; j++ ) {
		memcpy( (char *)&sample, in+j*4, 2 );
		i+=sample*sin( g*j );
		q+=sample*cos( g*j );
		id+=sample*sin( gd*j );
		qd+=sample*cos( gd*j );
		iu+=sample*sin( gu*j );
		qu+=sample*cos( gu*j );
	}
	t=sqrt( i*i+q*q );
	td=sqrt( id*id+qd*qd );
	tu=sqrt( iu*iu+qu*qu );
	printf( "t=%10.0f td=%10.0f tu=%10.0f", t, td, tu );
	if( td>t && tu<t ) { printf( " go up! " ); }
	if( tu>t && td<t ) { printf( " go dn! " ); }
	printf( "\n" );
}

void *audioFFT( void *a ) {

	workp works =(workp)a;
	double *co = NULL;
	double *si = NULL;
	double *fftr = NULL;
	double *ffti = NULL;
	double *coeff = NULL;
	int *fftout = NULL;
	int N;
	double co1, si1;
	double pi=3.141592653589793238462643383279502884197169399375105820974944592;
	double fftdr, fftdi;
	int j, k; 
	int m, p, md, ng, ig, i, ind, i1, i2, jd, kp;
	long val, valmax, freq;
	(void)freq;

	struct timeval start, stop;
	
	p=10;
	N=1<<p;

	fftr = calloc( N, sizeof( double ) );
	if( fftr == NULL ) {
		fprintf( stderr, "error allocating fftr buffer" );
		exit( 11 );
	}
	ffti = calloc( N, sizeof( double ) );
	if( ffti == NULL ) {
		fprintf( stderr, "error allocating ffti buffer" );
		exit( 12 );
	}
	fftout = calloc( N, sizeof( int ) );
	if( fftout == NULL ) {
		fprintf( stderr, "error allocating fftout buffer" );
		exit( 13 );
	}
	co = calloc( N/2, sizeof( double ) );
	if( co == NULL ) {
		fprintf( stderr, "error allocating co buffer" );
		exit( 14 );
	}
	si = calloc( N/2, sizeof( double ) );
	if( si == NULL ) {
		fprintf( stderr, "error allocating si buffer" );
		exit( 15 );
	}
	coeff = calloc( N, sizeof( double ) );
	if( coeff == NULL ) {
		fprintf( stderr, "error allocating coeff buffer" );
		exit( 16 );
	}

	// coeff table initialization
	for( j=1; j<N; j++ ) {
		coeff[j]=0.35875-0.48829*cos(2*pi*j/(N-1))+0.14128*cos(4*pi*j/(N-1))-0.01168*cos(6*pi*j/(N-1));
	}

	// sin table initialization
	m = N/2;
	p = 10;
	co1 = cos( pi/m );
	si1 = -sin( pi/m );
	co[0] = 1;
	si[0] = 0;
	for( j=1; j<m; j++ ) {
		co[j] = co1 * co[j-1] - si1 * si[j-1];
		si[j] = si1 * co[j-1] + co1 * si[j-1];
	}

	// continuous loop
	syslog( LOG_INFO, "audioFFT thread started" );
	prctl( PR_SET_NAME, (unsigned long)"AudioFFT", 0, 0, 0 );

	for( ; works->unterminated; ) {		
		pthread_cond_wait( works->audioFftCond,  works->audioFftMut );
		//compute( works->audioFftIn, sizeof(works->audioFftIn)/AUDIO_BYTES_PER_SAMPLE, works->serParms->pitch );

		// Time measurement
		gettimeofday( &start, NULL );
		
		// Read input data 
		for( j=0; j<N; j++ ) {
			fftr[j]=((short *)works->audioFftIn)[2*j]*coeff[j];
			//ffti[j]=((float *)works->audioFftIn)[2*j+1]*coeff[j];
			ffti[j]=0;
		}

		m = N/2;
		p = 10;
			
		// FFT Compute
		for( j=0,ng=1,md=m; j<p; j++ ) {
			for( ind=0, ig=0; ig<ng; ig++ ) {
				for( i=0, k=0; i<md; i++ ) {
					i1=ind+i;
					i2=i1+md;
					fftdr = fftr[i1] - fftr[i2];
					fftdi = ffti[i1] - ffti[i2];
					fftr[i1] = fftr[i1] + fftr[i2];
					ffti[i1] = ffti[i1] + ffti[i2];
					fftr[i2] = fftdr * co[k] - fftdi * si[k];
					ffti[i2] = fftdr * si[k] + fftdi * co[k];
					k += ng;
				}
				ind += md*2;
			}
			md >>= 1;
			ng <<= 1;
		}

		// Coefficients reordering
		for( j=0; j<N; j++ ) {
			jd=j;
			k=0;
			kp=m;
			for( i=1; i<=p; i++ ) {
				k = k + (jd-(jd/2)*2)*kp;
				jd >>= 1;
				kp >>= 1;
			}
			if( k>j ) {
				fftdr = fftr[j];
				fftdi = ffti[j];
				fftr[j] = fftr[k];
				ffti[j] = ffti[k];
				fftr[k] = fftdr;
				ffti[k] = fftdi;
			}
		}

		// Absolute value
		for( j=0, valmax=-200; j<N/8; j++ ) {
			fftr[j] = fftr[j]/m;
			ffti[j] = ffti[j]/m;
			fftr[j] = sqrt( fftr[j] * fftr[j] + ffti[j] * ffti[j] );
			val=20*log10(fftr[j]);
			fftout[j]=(int)val;
			if( val > valmax ) {
				valmax=val;
			}
		}
		for( j=0; j<N/8; j++ ) {
			if( fftout[j]==valmax ) {
				freq=j*48000/N;
			}
			fftout[j]+=200-valmax;
			if(  fftout[j] < 0 ) {
				 fftout[j]=0;
			}
		}
		// time measurement
		gettimeofday( &stop, NULL );
		//fprintf( stderr, "FFT time= %f mSec freq=%ld\n", (stop.tv_sec-start.tv_sec)*1000+0.001*(stop.tv_usec-start.tv_usec), freq );

		for( j=0; j<N; j++ ) {
			works->audioFftOut[j]=fftout[j];
		}
		//memcpy( works->audioFftOut, fftout, 512 );
		//memcpy( buf+10, fftout+512+64, 512-64 );
		//memcpy( buf+512+10-64, fftout, 512-64 );
	}
	
	syslog( LOG_INFO, "audioFFT operations end\n" );
	pthread_exit( a );
	return( a );
}



