/*
	Copyright 2018-2019 Giovanni Franza - HB9EIK 

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <syslog.h>
#include "struct.h"

static int baudRate[] = { 1200, 1800, 2400, 4800, 9600, 19200, 38400 };
static char *bwCW[]={ " ", "2600", "1500", "1000", "500", "300", "100", "100a", "100b", "100c", "100d" };
static char *bwAM[]={ " ", "2500", "3000", "3500", "4000", "4500", "5000", "5500", "6000" };
static char *bwFM[]={ " ", "Nar", "Wide", "Data" };
static char *bwSSB[]={ " ", "1600", "1700", "1800", "1900", "2000", "2100", "2200", "2300", "2400", "2500", "2600",
        "2700", "2800", "2900", "3000", "3100", "4000", "5000", "6000", "D300", "D600", "D1K" };
static char **filterDesc[]= { NULL, bwAM, bwSSB, bwSSB, bwCW, bwFM, bwCW };
static char *modes[]={ " ", "AM", "LSB", "USB", "CW+", "FM", "CW-" };
static int filterMode[]={ 0, 5, 1, 2, 3, 4, 3 };
static int filterMin[]={ 0, 0, 0, 0, 7, 0, 7 };
static int filterMax[]={ 0, 7,21,21,16, 2,16 };
static char *agcString[]={ " ", "Auto", "Man" };
static char *empty[]={ " ", " ", " " };
static char *fast[]={ " ", "Slow", "Med", "Fast", "!" };
static char *decodePaString[]={ "", "ByPass", "On", "Fault" };
static char *decodeTuString[]={ "", "ByPass", "On", "Tuning" };
static char *decodeFFTtype[]={ "DUO", "ART1k", "ART2k" };
static char *decodeAliased[]={ "", "Aliased", "Anti" };
static char *decodeGraphic[]={ "", "Draw", "Graphic" };
static char *decodeIsTouch[]={ "","Mouse", "Touch" };
static char *decodeFMdev[]= { "","2.5k", "5k" };
static char *decodeKePad[]={ "", "Key", "Paddle" };
static char *decodeMute[]={ "","Unmuted", "Muted" };
static char *decodeTXIn[]={ "Mic", "USB", "Auto" };
static char *decodeTXOut[]={ "","Normal", "0dBm" };
static char *decodeAntN[]={ "", "1", "2" };
static char *decodePwr[]={ "0.3 W", "0.5 W", "1.0 W", "1.2 W", "1.5 W", "2.0 W", "3.0 W", "4.0 W", "5.0 W", "Max" }; 
static char *decodeTXBW[]={ "50-4k", "100-2.7k", "100-3k", "100-3.5k", "100-4k", "200-2.7k", "200-3k", "200-3.5k", "200-4k", "300-2.7k", "300-3k", "300-3.5k", "300-4k" };
static long ncompress[]={ 0, 8000, 16000, 32000, 48000, 64000, 96000, 128000, 192000, 256000, 384000, 512000, 1024000 };
static char *scompress[]={ "Uncompr", "8k", "16k", "32k", "48k", "64k", "96k", "128k", "192k", "256k", "384k", "512k", "1M" };
static float nchunk[]={ 1, 2.5, 5, 10, 20 };
static char *schunk[]={ "1ms", "2.5ms", "5ms", "10ms", "20ms" };
static long nbtime[]={ 2, 5, 10, 20, 50, 100, 200, 500, 1000, 2000, 5000, 10000, 20000 };
static char *sbtime[]={ "2ms", "5ms", "10ms", "20ms", "50ms", "100ms", "200ms", "500ms", "1s", "2s", "5s", "10s", "20s" };
static char *decodeKeyP[]={ " ", "Key", "Paddle" };
static char *On[]={"On"};
static char *BWNULL[]={ " ", " ", " ", " ", " " };
static char *txAttStr[]={ "0dB" , "2dB", "3db", "5dB", "St-By" };
static int bwCWf[]={ 0, 1300, 750, 500, 250, 150, 50, 50, 50, 50, 50 };
static int bwCWi[]={ 0, -1300, -750, -500, -250, -150, -50, -50, -50, -50, -50 };
static int bwAMi[]={ 0, -2550, -3000, -3500, -4000, -4500, -5000, -5500, -6000 };
static int bwAMf[]={ 0, 2550, 3000, 3500, 4000, 4500, 5000, 5500, 6000 };
static int bwFMi[]={ 0, -6000, -6000, -11000 };
static int bwFMf[]={ 0, 6000, 6000, 11000 };
static int bwSSBi[]={ 0, +180, +180, +180, +180, +180, +180, +180, +180, +180, +180, +180,
        +180, +180, +180, +180, +180, +180, +180, +180, +1350, +1200, +1000 };
static int bwSSBf[]={ 0, 1600+180, 1700+180, 1800+180, 1900+180, 2000+180, 2100+180, 2200+180, 2300+180, 2400+180, 2500+180, 2600+180,
        2700+180, 2800+180, 2900+180, 3000+180, 3100+180, 4000+180, 5000+180, 6000+180, 1650, 1800, 2000 };
static char *error[]= { "none", "1-High temperature", "HI SWR error", "1-High temperature", "PWR IN error",
        "Comm warn ", "Remote warn Usb", "Audio not found", "DUO mode not activated", "DUO EXTIO warning",
        "Audio read warning", "Audio write warning", "Web socket warning", "Auto Tune timeout", "Auto Tune SWR warning",
        "Auto Tune DUO aborts", "Error ST", "Micro Reset warning", "CMD not available", "Auto Tune Max SWR",
        "Band warn StdBy", "Remote warn Cat", "2-High temperature", "2-High temperature", "2-Warning temperature",
        "2-Warning temperature", "Error Open serial", "Error Socket Close", "Auto Tune Low Pwr" };
static char *status[] = { "Generic", "FDM DUO", "FT817" };
static char *decodeAB[]={ "", "\"A\"","\"B\""};
static char *decodeABn[]={ "", "A","B"};
static char *messages[]={ "VFO %s: ", "Band:", "Wired for:", "Temperature: %4.1f °C", "Voltage:", "Current:", "IP:", "IRTT:", "in:", "out:", "max", "med", "A", "V,", "ms", "Error:", "RTX IP:", "(Kbit/s)",
        "WPM", "Max", "mW", "s", "dB", "Gain:", "Step:", "St:", "KHz", "Hz", "S:", "MHz", "A", "M", "AGC=", "Speed=", "Off", "NB=", "NR=", "AN=", "Rit:", "SWR", "Frequency", "W", "Fwd", "Ref", "Gain=", "Thr=", "PA Out", "dbm", "Undecoded - unrecognized",
	"TmF1", "TmF2", "TmF3", "TmF4", "TmF5", "TmF6", "TmE1", "TmE2", "TmE3", "TmE1", "TmE2", "TmE3", "KbF1", "KbF2", "KbF3", "KbF4", "KbF5", "KbF6", "KbF7", "KbF8", "KbF9", "KbF10", "KbF11", "KbF12", ".....Off" };

static char a1[11][30];
static char *SbwCW[]={a1[0],a1[1],a1[2],a1[3],a1[4],a1[5],a1[6],a1[7],a1[8],a1[9],a1[10]};
static char a2[9][30];
static char *SbwAM[]={a2[0],a2[1],a2[2],a2[3],a2[4],a2[5],a2[6],a2[7],a2[8]};
static char a3[4][30];
static char *SbwFM[]={a3[0],a3[1],a3[2],a3[3]};
static char a4[23][30];
static char *SbwSSB[]={a4[0],a4[1],a4[2],a4[3],a4[4],a4[5],a4[6],a4[7],a4[8],a4[9],a4[10],a4[11],a4[12],a4[13],a4[14],a4[15],a4[16],a4[17],a4[18],a4[19],a4[20],a4[21],a4[22]};
static char a5[7][30];
static char *Smodes[]={a5[0],a5[1],a5[2],a5[3],a5[4],a5[5],a5[6]};
static char a6[3][30];
static char *SagcString[]={a6[0],a6[1],a6[2]};
static char a7[4][30];
static char *Sempty[]={a7[0],a7[1],a7[2],a7[3]};
static char a8[4][30];
static char *Sfast[]={a8[0],a8[1],a8[2],a8[3]};
static char a9[4][30];
static char *SdecodePaString[]={a9[0],a9[1],a9[2],a9[3]};
static char a10[4][30];
static char *SdecodeTuString[]={a10[0],a10[1],a10[2],a10[3]};
static char a11[3][30];
static char *SdecodeFFTtype[]={a11[0],a11[1],a11[2]};
static char a12[3][30];
static char *SdecodeAliased[]={a12[0],a12[1],a12[2]};
static char a13[3][30];
static char *SdecodeGraphic[]={a13[0],a13[1],a13[2]};
static char a14[3][30];
static char *SdecodeIsTouch[]={a14[0],a14[1],a14[2]};
static char a15[3][30];
static char *SdecodeFMdev[]={a15[0],a15[1],a15[2]};
static char a16[3][30];
static char *SdecodeKePad[]={a16[0],a16[1],a16[2]};
static char a17[3][30];
static char *SdecodeMute[]={a17[0],a17[1],a17[2]};
static char a18[3][30];
static char *SdecodeTXIn[]={a18[0],a18[1],a18[2]};
static char a19[3][30];
static char *SdecodeTXOut[]={a19[0],a19[1],a19[2]};
static char a20[3][30];
static char *SdecodeAntN[]={a20[0],a20[1],a20[2]};
static char a21[10][30];
static char *SdecodePwr[]={a21[0],a21[1],a21[2],a21[3],a21[4],a21[5],a21[6],a21[7],a21[8],a21[9]};
static char a22[13][30];
static char *SdecodeTXBW[]={a22[0],a22[1],a22[2],a22[3],a22[4],a22[5],a22[6],a22[7],a22[8],a22[9],a22[10],a22[11],a22[12]};
static char a23[13][30];
static char *Sscompress[]={a23[0],a23[1],a23[2],a23[3],a23[4],a23[5],a23[6],a23[7],a23[8],a23[9],a23[10],a23[11],a23[12]};
static char a24[5][30];
static char *Sschunk[]={a24[0],a24[1],a24[2],a24[3],a24[4]};
static char a25[13][30];
static char *Ssbtime[]={a25[0],a25[1],a25[2],a25[3],a25[4],a25[5],a25[6],a25[7],a25[8],a25[9],a25[10],a25[11], a25[12]};
static char a26[3][30];
static char *SdecodeKeyP[]={a26[0],a26[1],a26[2]};
static char a27[2][30];
static char *SOn[]={a27[0], a27[1]};
static char a28[5][30];
static char *SBWNULL[]={a28[0],a28[1],a28[2],a28[3],a28[4]};
static char a29[5][30];
static char *StxAttStr[]={a29[0],a29[1],a29[2],a29[3],a29[4]};
static char a30[28][30];
static char *Serror[]={a30[0],a30[1],a30[2],a30[3],a30[4],a30[5],a30[6],a30[7],a30[8],a30[9],a30[10],a30[11],a30[12],a30[13],a30[14],a30[15],a30[16],a30[17],a30[18],a30[19],a30[20],a30[21],a30[22],a30[23],a30[24],a30[25],a30[26],a30[27]};
static char a31[4][30];
static char *Sstatus[]={a31[0],a31[1],a31[2],a31[3]};
static char a32[3][30];
static char *SdecodeAB[]={a32[0],a32[1],a32[2]};
static char a33[3][30];
static char *SdecodeABn[]={a33[0],a33[1],a33[2]};
static char a34[74][30];
static char *Smessages[]={ a34[0], a34[1], a34[2], a34[3], a34[4], a34[5], a34[6], a34[7], a34[8], a34[9], a34[10], a34[11], a34[12], a34[13], a34[14], a34[15], a34[16], a34[17], a34[18], a34[19], a34[20], a34[21], a34[22], a34[23], a34[24], a34[25], a34[26], a34[27], a34[28], a34[29], a34[30], a34[31], a34[32], a34[33], a34[34], a34[35], a34[36], a34[37], a34[38], a34[39], a34[40], a34[41], a34[42], a34[43], a34[44], a34[45], a34[46], a34[47], a34[48], a34[49], a34[50], a34[51], a34[52], a34[53], a34[54], a34[55], a34[56], a34[57], a34[58], a34[59], a34[60], a34[61], a34[62], a34[63], a34[64], a34[65], a34[66], a34[67], a34[68], a34[69], a34[70], a34[71], a34[72], a34[73] };
static char **SfilterDesc[]= { NULL, SbwAM, SbwSSB, SbwSSB, SbwCW, SbwFM, SbwCW };

void locale( workp works, char *langue ) {
	
	FILE *fd;
	char buf[1000];
	char *pp, *qq;
	char ***what;
	int index;
	int isButton=0;
	int lun;

	works->strings->baudRate = baudRate;
	works->strings->nBaudRate = sizeof(baudRate)/sizeof(baudRate[0]);

	if( langue==NULL ) {
		works->strings->bwCW = bwCW;
		works->strings->bwAM = bwAM;
		works->strings->bwFM = bwFM;
		works->strings->bwSSB = bwSSB;
		works->strings->empty = empty;
		works->strings->modes = modes;
		works->strings->filterDesc = filterDesc;
		works->strings->filterMode = filterMode;
		works->strings->filterMin = filterMin;
		works->strings->filterMax = filterMax;
		works->strings->agcString = agcString;
		works->strings->fast = fast;
		works->strings->decodePaString = decodePaString;
		works->strings->decodeTuString = decodeTuString;
		works->strings->decodeFFTtype = decodeFFTtype;
		works->strings->decodeAliased = decodeAliased;
		works->strings->decodeGraphic = decodeGraphic;
		works->strings->decodeIsTouch = decodeIsTouch;
		works->strings->decodeFMdev = decodeFMdev;
		works->strings->decodeKePad = decodeKePad;
		works->strings->decodeMute = decodeMute;
		works->strings->decodeTXIn = decodeTXIn;
		works->strings->decodeTXOut = decodeTXOut;
		works->strings->decodeAntN = decodeAntN;
		works->strings->decodePwr = decodePwr;
		works->strings->decodeTXBW = decodeTXBW;
		works->strings->ncompress = ncompress;
		works->strings->scompress = scompress;
		works->strings->nchunk = nchunk;
		works->strings->schunk = schunk;
		works->strings->nbtime = nbtime;
		works->strings->sbtime = sbtime;
		works->strings->decodeKeyP = decodeKeyP;
		works->strings->On = On;
		works->strings->BWNULL = BWNULL;
		works->strings->txAttStr = txAttStr;
		works->strings->bwCWf=bwCWf;
		works->strings->bwCWi=bwCWi;
		works->strings->bwAMi=bwAMi;
		works->strings->bwAMf=bwAMf;
		works->strings->bwFMi=bwFMi;
		works->strings->bwFMf=bwFMf;
		works->strings->bwSSBi=bwSSBi;
		works->strings->bwSSBf=bwSSBf;
		works->strings->error=error;
		works->strings->status=status;
		works->strings->decodeAB=decodeAB;
		works->strings->decodeABn=decodeABn;
		works->strings->messages=messages;
		works->strings->nerror=XtNumber(error);
		works->strings->nstatus=XtNumber(status);
		return;
	}
	sprintf( buf, "langue-%s.txt", langue );
        fd = fopen( buf, "r" );
        if ( !fd ) {
                syslog( LOG_ERR, "can't open %s", buf );
                return;
        }
        for( ;; ) {
                pp = fgets( buf, sizeof( buf ), fd );
                if( !pp ) {
                        break;
                }
		if( pp[0]=='#' ) {
			continue;
		}
		if( pp[strlen(pp)-1]=='\n' ) {
			pp[strlen(pp)-1]=0;
		}
		if( pp[strlen(pp)-1]=='\r' ) {
			pp[strlen(pp)-1]=0;
		} 
		if( pp[strlen(pp)-1]=='\n' ) {
			pp[strlen(pp)-1]=0;
		}
		if( pp[0]=='$' ) {
			if( !strcmp( pp+1, "bwCW" ) ) {
				what=&works->strings->bwCW;
				*what=SbwCW;
				lun=XtNumber(SbwCW);
			} else if( !strcmp( pp+1, "bwAM" ) ) {
				what=&works->strings->bwAM;
				*what=SbwAM;
				lun=XtNumber(SbwAM);
			} else if( !strcmp( pp+1, "bwFM" ) ) {
				what=&works->strings->bwFM;
				*what=SbwFM;
				lun=XtNumber(SbwFM);
			} else if( !strcmp( pp+1, "bwSSB" ) ) {
				what=&works->strings->bwSSB;
				*what=SbwSSB;
				lun=XtNumber(SbwSSB);
			} else if( !strcmp( pp+1, "empty" ) ) {
				what=&works->strings->empty;
				*what=Sempty;
				lun=XtNumber(Sempty);
			} else if( !strcmp( pp+1, "modes" ) ) {
				what=&works->strings->modes;
				*what=Smodes;
				lun=XtNumber(Smodes);
			} else if( !strcmp( pp+1, "agcString" ) ) {
				what=&works->strings->agcString;
				*what=SagcString;
				lun=XtNumber(SagcString);
			} else if( !strcmp( pp+1, "fast" ) ) {
				what=&works->strings->fast;
				*what=Sfast;
				lun=XtNumber(Sfast);
			} else if( !strcmp( pp+1, "decodePaString" ) ) {
				what=&works->strings->decodePaString;
				*what=SdecodePaString;
				lun=XtNumber(SdecodePaString);
			} else if( !strcmp( pp+1, "decodeTuString" ) ) {
				what=&works->strings->decodeTuString;
				*what=SdecodeTuString;
				lun=XtNumber(SdecodeTuString);
			} else if( !strcmp( pp+1, "decodeFFTtype" ) ) {
				what=&works->strings->decodeFFTtype;
				*what=SdecodeFFTtype;
				lun=XtNumber(SdecodeFFTtype);
			} else if( !strcmp( pp+1, "decodeAliased" ) ) {
				what=&works->strings->decodeAliased;
				*what=SdecodeAliased;
				lun=XtNumber(SdecodeAliased);
			} else if( !strcmp( pp+1, "decodeGraphic" ) ) {
				what=&works->strings->decodeGraphic;
				*what=SdecodeGraphic;
				lun=XtNumber(SdecodeGraphic);
			} else if( !strcmp( pp+1, "decodeIsTouch" ) ) {
				what=&works->strings->decodeIsTouch;
				*what=SdecodeIsTouch;
				lun=XtNumber(SdecodeIsTouch);
			} else if( !strcmp( pp+1, "decodeFMdev" ) ) {
				what=&works->strings->decodeFMdev;
				*what=SdecodeFMdev;
				lun=XtNumber(SdecodeFMdev);
			} else if( !strcmp( pp+1, "decodeKePad" ) ) {
				what=&works->strings->decodeKePad;
				*what=SdecodeKePad;
				lun=XtNumber(SdecodeKePad);
			} else if( !strcmp( pp+1, "decodeMute" ) ) {
				what=&works->strings->decodeMute;
				*what=SdecodeMute;
				lun=XtNumber(SdecodeMute);
			} else if( !strcmp( pp+1, "decodeTXIn" ) ) {
				what=&works->strings->decodeTXIn;
				*what=SdecodeTXIn;
				lun=XtNumber(SdecodeTXIn);
			} else if( !strcmp( pp+1, "decodeTXOut" ) ) {
				what=&works->strings->decodeTXOut;
				*what=SdecodeTXOut;
				lun=XtNumber(SdecodeTXOut);
			} else if( !strcmp( pp+1, "decodeAntN" ) ) {
				what=&works->strings->decodeAntN;
				*what=SdecodeAntN;
				lun=XtNumber(SdecodeAntN);
			} else if( !strcmp( pp+1, "decodePwr" ) ) {
				what=&works->strings->decodePwr;
				*what=SdecodePwr;
				lun=XtNumber(SdecodePwr);
			} else if( !strcmp( pp+1, "decodeTXBW" ) ) {
				what=&works->strings->decodeTXBW;
				*what=SdecodeTXBW;
				lun=XtNumber(SdecodeTXBW);
			} else if( !strcmp( pp+1, "scompress" ) ) {
				what=&works->strings->scompress;
				*what=Sscompress;
				lun=XtNumber(Sscompress);
			} else if( !strcmp( pp+1, "schunk" ) ) {
				what=&works->strings->schunk;
				*what=Sschunk;
				lun=XtNumber(Sschunk);
			} else if( !strcmp( pp+1, "sbtime" ) ) {
				what=&works->strings->sbtime;
				*what=Ssbtime;
				lun=XtNumber(Ssbtime);
			} else if( !strcmp( pp+1, "decodeKeyP" ) ) {
				what=&works->strings->decodeKeyP;
				*what=SdecodeKeyP;
				lun=XtNumber(SdecodeKeyP);
			} else if( !strcmp( pp+1, "On" ) ) {
				what=&works->strings->On;
				*what=SOn;
				lun=XtNumber(SOn);
			} else if( !strcmp( pp+1, "BWNULL" ) ) {
				what=&works->strings->BWNULL;
				*what=SBWNULL;
				lun=XtNumber(SBWNULL);
			} else if( !strcmp( pp+1, "txAttStr" ) ) {
				what=&works->strings->txAttStr;
				*what=StxAttStr;
				lun=XtNumber(StxAttStr);
			} else if( !strcmp( pp+1, "error" ) ) {
				what=&works->strings->error;
				*what=Serror;
				lun=XtNumber(Serror);
			} else if( !strcmp( pp+1, "status" ) ) {
				what=&works->strings->status;
				*what=Sstatus;
				lun=XtNumber(Sstatus);
			} else if( !strcmp( pp+1, "AB" ) ) {
				what=&works->strings->decodeAB;
				*what=SdecodeAB;
				lun=XtNumber(SdecodeAB);
			} else if( !strcmp( pp+1, "ABn" ) ) {
				what=&works->strings->decodeABn;
				*what=SdecodeABn;
				lun=XtNumber(SdecodeABn);
			} else if( !strcmp( pp+1, "messages" ) ) {
				what=&works->strings->messages;
				*what=Smessages;
				lun=XtNumber(Smessages);
			}
			if( !strcmp( pp+1, "BUTTONS" ) ) {
				isButton=1;
			} else {
				isButton=0;
			}
		} else {
			if( pp[3]==';' ) {
				pp[3]='\0';
				index=atoi( pp );
				if( isButton ) {
					if( index > 0 && index <= works->nbuttons ) {
						if( (works->buttons+index-1)->name ) {
							free( (works->buttons+index-1)->name );
						}
						qq = calloc( strlen( pp+4 )+1, 1 );
						if( !qq ) {
							syslog( LOG_ERR, "locale buttons name calloc (%d - %s) failed", index, pp+4 );
							exit( 1 );
						}
						strcpy( qq, pp+4 );
						(works->buttons+index-1)->name=qq;
					}
				} else {
					if( index < lun && index>=0 ) {
						strcpy( (*what)[index], pp+4 );
					}
				}
			}
		}
	}
	works->strings->filterDesc = SfilterDesc;
}

