/*
	Copyright 2018-2019 Giovanni Franza - HB9EIK 

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <math.h>
#include <syslog.h>
#include <unistd.h>
#include <dlfcn.h>
#include <portaudio.h>
#include "struct.h"

#define SAMPLE_RATE         (48000)
#define PA_SAMPLE_TYPE      paInt16

typedef float SAMPLE;

static int outCallback( const void *inputBuffer, void *outputBuffer, unsigned long framesPerBuffer, const PaStreamCallbackTimeInfo* timeInfo, PaStreamCallbackFlags statusFlags, void *userData );

void MyError(const char *file, int line, const char *function, int err, const char *fmt,...) {
	return;
}
static int outCallback( const void *inputBuffer, void *outputBuffer, unsigned long framesPerBuffer, const PaStreamCallbackTimeInfo* timeInfo, PaStreamCallbackFlags statusFlags, void *userData ) {
	static int count=0;
	SAMPLE *out = (SAMPLE*)outputBuffer;
	(void) inputBuffer;
	workp works = (workp) userData;
	(void) timeInfo; /* Prevent unused variable warnings. */
	(void) statusFlags;
	int j;
	int k;
	int toBeCopied;
	int divider = works->rx_buffer_frames*works->buffer_start*AUDIO_BYTES_PER_SAMPLE*2; // buffer twice the start buffer
	if( out && !works->shutAudio && works->soundConnected ) {
		for( j=0, k=0; j<framesPerBuffer*2; j++, k++ ) {
			((short *)out)[k]=((short *)(works->sound_buffer+works->soundReadPointer))[j]*works->setup->auxVol/100;
			if( works->setup->out_channels==1 ) {
				j++;
			}
		}
		if( count >=0 ) {
			toBeCopied = count*framesPerBuffer*AUDIO_BYTES_PER_SAMPLE;
			if( toBeCopied<sizeof(works->audioFftIn) ) {
				memcpy( works->audioFftIn+toBeCopied, (works->sound_buffer+works->soundReadPointer), sizeof(works->audioFftIn)-toBeCopied > framesPerBuffer*AUDIO_BYTES_PER_SAMPLE ?  framesPerBuffer*AUDIO_BYTES_PER_SAMPLE : sizeof(works->audioFftIn)-toBeCopied  );
				count++;
			} else {
				pthread_cond_signal( works->audioFftCond );
				count=FFT_AUDIO_DECIM;
			}
		} else {
			count++;
		}
		memset( (works->sound_buffer+works->soundReadPointer), 0, framesPerBuffer*AUDIO_BYTES_PER_SAMPLE );
		works->soundReadPointer+=framesPerBuffer*AUDIO_BYTES_PER_SAMPLE;
		works->soundReadPointer%=divider;
	} else {
		memset( (char *)out, 0, framesPerBuffer*AUDIO_BYTES_PER_SAMPLE );
	}
	if( works->bufferingChanged ) {
		syslog( LOG_INFO, "buffer changed" );
		works->bufferingChanged = 0;
		works->soundWritePointer = works->buffer_start;
		works->soundReadPointer = 0;
	}
	if( !works->soundReadPointer && !works->shutAudio && works->soundConnected ) {
		if( count < FFT_AUDIO_DECIM/3 && count > 2*FFT_AUDIO_DECIM/3 ) {
			if( works->soundWritePointer >= works->buffer_start*2 ) {
				syslog( LOG_INFO, "too much sound %d", works->buffer_start*2 );
				works->soundWritePointer = works->buffer_start;
				count=FFT_AUDIO_DECIM;
			} else if ( works->soundWritePointer <= 0 ) { 
				syslog( LOG_INFO, "sound exausted" );
				works->soundWritePointer = works->buffer_start;
				count=FFT_AUDIO_DECIM;
			}
		}
	}
	return paContinue;
}

void *th_audio( void *parms ) {
	PaStreamParameters inputParameters, outputParameters;
	PaStream *streamw;
	PaStream *streamr;
	PaError err;
	void *dll;
	void (*func)( void * );
	workp works = (workp)parms;
	int ndev;
	char myBuffer[20480];
	int j;

	prctl( PR_SET_NAME, (unsigned long)"th_audio", 0, 0, 0 );

	dll = dlopen( "libasound.so", RTLD_LAZY);
	if( dll ) {
		dlerror();
		func = dlsym( dll, "snd_lib_error_set_handler" );
		if( dlerror() == NULL ) {
			func( MyError );
		}
	}

	err = Pa_Initialize();
	if( err != paNoError ) {
		Pa_Terminate();
		syslog( LOG_ERR, "An error occured while using the portaudio stream: %d - %s", err, Pa_GetErrorText( err ) );
		works->unterminated=0;
		return NULL;
		// continue;
	}

	ndev = Pa_GetDeviceCount( );
	if( ndev < 1 ) {
		Pa_Terminate();
		syslog( LOG_ERR, "An error occured while using the portaudio stream: no devices" );
		works->unterminated=0;
		return NULL;
		//continue;
	}
	
	for( ;; ) {
		usleep( 100000 );
		if( works->noAudio ) {
			continue;
		}

		if( works->setup->devTx >= ndev ) {
			Pa_Terminate();
			syslog( LOG_ERR, "An error occured while using the portaudio stream: no input device %d", works->setup->devTx );
			//works->unterminated=0;
			//return NULL;
			continue;
		}

		inputParameters.device = works->setup->devTx;
		if( inputParameters.device == -1 ) {
			inputParameters.device = Pa_GetDefaultInputDevice();
			if (inputParameters.device == paNoDevice) {
				Pa_Terminate();
				syslog( LOG_ERR, "An error occured while using the portaudio stream: no input device %d", inputParameters.device );
				//works->unterminated=0;
				//return NULL;
				continue;
			}
		}

		//works->setup->devTx = inputParameters.device;
		inputParameters.channelCount = works->setup->in_channels;
		inputParameters.sampleFormat = PA_SAMPLE_TYPE;
		inputParameters.suggestedLatency = Pa_GetDeviceInfo( inputParameters.device )->defaultLowInputLatency;
		inputParameters.hostApiSpecificStreamInfo = NULL;

		if( works->setup->devRx >= ndev ) {
			Pa_Terminate();
			syslog( LOG_ERR, "An error occured while using the portaudio stream: no output device %d", works->setup->devRx );
			//works->unterminated=0;
			//return NULL;
			continue;
		}

		outputParameters.device = works->setup->devRx;
		if( outputParameters.device == -1 ) {
			outputParameters.device = Pa_GetDefaultOutputDevice();
			if (outputParameters.device == paNoDevice) {
				Pa_Terminate();
				syslog( LOG_ERR, "An error occured while using the portaudio stream: no output device %d", outputParameters.device );
				//works->unterminated=0;
				//return NULL;
				continue;
			}
		}

		//works->setup->devRx = outputParameters.device;
		outputParameters.channelCount = works->setup->out_channels;
		outputParameters.sampleFormat = PA_SAMPLE_TYPE;
		outputParameters.suggestedLatency = Pa_GetDeviceInfo( outputParameters.device )->defaultLowOutputLatency;
		// outputParameters.suggestedLatency = 0.005;
		// printf( "suggestedLatency=%f\n", outputParameters.suggestedLatency );
		outputParameters.hostApiSpecificStreamInfo = NULL;

		err = Pa_OpenStream( &streamw, NULL, &outputParameters, SAMPLE_RATE, works->setup->AudioBuffers, 0, outCallback, parms );
		if( err != paNoError ) {
			Pa_Terminate();
			syslog( LOG_ERR, "An error occured while opening the portaudio output stream: %d - %s", err, Pa_GetErrorText( err ) );
			//works->unterminated=0;
			//return NULL;
			continue;
		}

		err = Pa_OpenStream( &streamr, &inputParameters, NULL, SAMPLE_RATE, works->setup->AudioBuffers, 0, NULL, parms );
		if( err != paNoError ) {
			Pa_Terminate();
			syslog( LOG_ERR, "An error occured while opening the portaudio input stream: %d - %s", err, Pa_GetErrorText( err ) );
			//works->unterminated=0;
			//return NULL;
			continue;
		}

		err = Pa_StartStream( streamw );
		if( err != paNoError ) {
			Pa_Terminate();
			syslog( LOG_ERR, "An error occured while starting the portaudio output stream: %d - %s", err, Pa_GetErrorText( err ) );
			//works->unterminated=0;
			//return NULL;
			continue;
		}

		err = Pa_StartStream( streamr );
		if( err != paNoError ) {
			Pa_Terminate();
			syslog( LOG_ERR, "An error occured while starting the portaudio input stream: %d - %s", err, Pa_GetErrorText( err ) );
			//works->unterminated=0;
			//return NULL;
			continue;
		}

		works->started=2;
		for( ; works->unterminated && works->noAudio==0 ; ) {
			err = Pa_ReadStream( streamr, myBuffer, works->tx_buffer_frames );
			if( works->setup->in_channels==1 ) {
				for( j=works->tx_buffer_frames*2-2; j>0; j-=2 ) {
					memcpy( myBuffer+j*2, myBuffer+j, 2 );
					memcpy( myBuffer+j*2+2, myBuffer+j, 2 );
				}
			}
			if( works->txCwLock==0 && works->serParms->mode != 4 && works->serParms->mode != 6 ) {
				sendWebSockAudioFrame( works, myBuffer );
			}
		}

		err = Pa_StopStream( streamw );
		if( err != paNoError ) {
			Pa_Terminate();
			syslog( LOG_ERR, "An error occured while stopping the portaudio output stream: %d - %s", err, Pa_GetErrorText( err ) );
			works->unterminated=0;
			return NULL;
		}

		err = Pa_StopStream( streamr );
		if( err != paNoError ) {
			Pa_Terminate();
			syslog( LOG_ERR, "An error occured while stopping the portaudio input stream: %d - %s", err, Pa_GetErrorText( err ) );
			works->unterminated=0;
			return NULL;
		}

		works->started=2;
		err = Pa_CloseStream( streamw );
		if( err != paNoError ) {
			Pa_Terminate();
			syslog( LOG_ERR, "An error occured while closing the portaudio output stream: %d - %s", err, Pa_GetErrorText( err ) );
			works->unterminated=0;
			return NULL;
		}

		err = Pa_CloseStream( streamr );
		if( err != paNoError ) {
			Pa_Terminate();
			syslog( LOG_ERR, "An error occured while closing the portaudio input stream: %d - %s", err, Pa_GetErrorText( err ) );
			works->unterminated=0;
			return NULL;
		}

		if( !works->unterminated ) {
			break;
		}
		works->started=0;
	}
	Pa_Terminate();
	return NULL;
}
