/*
	Copyright 2018-2019 Giovanni Franza - HB9EIK 

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <openssl/sha.h>
#include <glib.h>
#include <syslog.h>

#include "struct.h"

static char *connectionHeader[] = {
	"GET /chat HTTP/1.1\r\n",
	"Host: %s:%d\r\n",
	"Upgrade: websocket\r\n",
	"Connection: keep-alive, Upgrade\r\n",
	"Sec-WebSocket-Key: %s\r\n",
	"Sec-WebSocket-Version: 13\r\n\r\n"
	"\r\n"
};

static char *responseHeader[] = {
	"HTTP/1.1 101 Switching Protocols",
	"Upgrade: websocket",
	"Connection: Upgrade",
	"Sec-WebSocket-Accept:",
	"Sec-WebSocket-Protocol: chat",
};
static char *recKeyHeader="Sec-WebSocket-Accept: ";

void *Wclient( void *a ) {
	int sfd;
	FILE *bsfd;
	struct sockaddr_in saddr;
	struct hostent *server;
	int j, k, n;
	int pos;
	int rc;
	short spos;
	int length;
	char mybuf[1024];
	char catbuf[200];
	opus_int16 convbuf[51200];
	char *pp;
	int status;
	char keys[100];
	char recKeys[100];
	workp works=(workp)a;
	int port = works->port;
	unsigned char SPbuf[4*1024];
	fd_set rset, wset;
	int flags;
	struct timeval tval, tval1, start, stop, rec;
	// struct timeval sstart, sstop;
	struct timeval FFTstart, FFTstop;
	int isFirst;
	int ra=1;
	int myerrno;
	long ns1, ns2, nus1, nus2;
	long s1, s2, us1, us2;
	int errnum;

	syslog( LOG_INFO, "Wclient to server %s on port %d mode %c", works->server, works->port, works->setup->webSocket );
	prctl( PR_SET_NAME, (unsigned long)"Wclient", 0, 0, 0 );

	sfd=open( "/dev/urandom", O_RDONLY );
	if( sfd == -1 ) {
		syslog( LOG_ERR, "/dev/urandom error %d", errno );
		return( NULL );
	}
	status = read( sfd, keys, 16 );
	if( status != 16 ) {
		syslog( LOG_ERR, "/dev/urandom read %d instead of 16, error=%d", status, errno );
		return( NULL );
	}
	close( sfd );
	pp=g_base64_encode( (const guchar *)keys, strlen(keys) );
	memset( keys, 0, sizeof( keys ) );
	strcpy( keys, (const char *)pp );
	g_free( pp );

	works->connected = 0;

	for( ;; ) {
		memset( works->sound_buffer, 0, sizeof( works->sound_buffer ) );
		strcpy( works->message, "Not connected to remote RTX" );
		works->soundConnected=0;
		works->countSo=0;
		works->countTo=0;
		works->countOldSo=-1;
		// socket: create the socket 
		works->sfd = socket( AF_INET, SOCK_STREAM, 0 );
		if( works->sfd == -1 ) {
			syslog( LOG_ERR, "Socket error %d", errno );
			return( NULL );
		}
		syslog( LOG_INFO, "Socket for TCP created" );

		rc=setsockopt(works->sfd, SOL_SOCKET, SO_REUSEADDR, (const char*)&ra, sizeof ra);
		if( rc == -1 ) {
			syslog( LOG_ERR, "Socket error %d %d", rc, errno );
		}
		syslog( LOG_INFO, "Socket for TCP reused" );

		flags = fcntl( works->sfd, F_GETFL, 0 );
		fcntl( works->sfd, F_SETFL, flags | O_NONBLOCK );
		server = gethostbyname( works->server );
		if (server == NULL) {
			syslog( LOG_ERR, "ERROR, no such host as %s\n", works->server );
			return( NULL );
		}
		syslog( LOG_INFO, "server ok" );

		memset( &saddr, 0, sizeof(saddr) );
		saddr.sin_family = AF_INET;
		saddr.sin_addr.s_addr = htonl( INADDR_ANY );
		memmove( (char *)&saddr.sin_addr.s_addr, (char *)server->h_addr, server->h_length);
		saddr.sin_port = htons( (unsigned short)port );
		syslog( LOG_INFO, "TCP bound" );

		/* connect: create a connection with the server */
		rc=connect( works->sfd, (struct sockaddr *)&saddr, sizeof( saddr ) );
		if( rc == -1 && errno != EINPROGRESS ) {
			syslog( LOG_ERR, "Connect Error" );
			works->connected=0;
			close( works->sfd );
			usleep( 10000 );
			continue;
		}
		if( rc ) { // 115 - EINPROGRESS
			FD_ZERO(&rset);
			FD_SET( works->sfd, &rset);
			wset = rset;
			for( ;; ) {
				tval.tv_sec = works->setup->wsockTimeout/1000;
				tval.tv_usec = (works->setup->wsockTimeout%1000)*1000;
				rc = select( works->sfd + 1, &rset, &wset, NULL, &tval );
				if ( rc != 1 ) {
					close( works->sfd );
					rc=(-1);
					break;
				}
				rc=0;
				break;
			}
			if( rc == -1 ) {
				continue;
			}
		}
		// now blocking
		rc=fcntl( works->sfd, F_SETFL, flags );

		// but, with timeout
		tval1.tv_sec = works->setup->wsockTimeout/1000;
		tval1.tv_usec = (works->setup->wsockTimeout%1000)*1000;
		rc=setsockopt(works->sfd, SOL_SOCKET, SO_RCVTIMEO, (const char*)&tval1, sizeof tval);
		rc=setsockopt(works->sfd, SOL_SOCKET, SO_SNDTIMEO, (const char*)&tval1, sizeof tval);

		// bsfd for buffered reading where needed
		bsfd=fdopen( works->sfd, "r+" );
		syslog( LOG_INFO, "TCP connected\n" );

		// websocket handshake
		for( j=0; j<XtNumber( connectionHeader ); j++ ) {
			if( j==1 ) {
				sprintf( mybuf, connectionHeader[j], works->server, works->port );
			} else if( j==4 ) {
				sprintf( mybuf, connectionHeader[4], keys );
			} else {
				strcpy( mybuf, connectionHeader[j] );
			}
			syslog( LOG_INFO, "TCP to write %d bytes (%-*.*s) \n", (int)strlen( mybuf ), (int)strlen( mybuf ), (int)strlen( mybuf ), mybuf );
			n = write( works->sfd, mybuf, strlen( mybuf ) );
			usleep( 50000 );
			if( n == -1 ) {
				syslog( LOG_ERR, "TCP write Error" );
				break;
			}
			syslog( LOG_INFO, "TCP write %d bytes\n", n );
		}
		syslog( LOG_INFO, "TCP headers writed\n" );
		errnum=0;
		for( status=0,pp=NULL;;pp=NULL ) {
			memset( mybuf, 0, sizeof( mybuf ) );
			if( pp==NULL ) {
				usleep( 10000 );
				syslog( LOG_INFO, "Reading TCP headers\n" );
				pp = fgets( mybuf, sizeof(mybuf), bsfd );
			}
			if( pp ) {
				syslog( LOG_INFO, "TCP read %d bytes", (int)strlen( pp ) );
			}
			if( !pp && errno==115 )  {
				syslog( LOG_INFO, "TCP no read" );
				errnum++;
				if( errnum > 10 ) {
					syslog( LOG_INFO, "TCP too many no read" );
					status=(-1);
					break;
				}
				continue;
			}
			if( !pp && errno )  {
				syslog( LOG_ERR, "TCP read error 1 %d", errno );
				status=(-1);
				break;
			}
			if( !pp ) {
				syslog( LOG_INFO, "TCP read 0 bytes" );
				continue;
			}
			errnum=0;
			if( mybuf[strlen(mybuf)-1]=='\n' ) { // buffer is long enough to have room for all an header
				syslog( LOG_INFO, "Received %d bytes %s", (int)strlen( mybuf ), mybuf );
				for( j=0; j<XtNumber(responseHeader); j++ ) {
					if( !strncmp( mybuf, responseHeader[j], strlen(responseHeader[j] ) ) ) {
						status++;
						syslog( LOG_INFO, "Caught %s go to status %d/%d\n", responseHeader[j], status, (int)(XtNumber(responseHeader)) );
						if( !strncmp( mybuf, recKeyHeader, strlen(recKeyHeader ) ) ) {
							syslog( LOG_INFO, "Caught key %s", mybuf+strlen(recKeyHeader) );
							strcpy( recKeys, mybuf+strlen(recKeyHeader) );
							if( recKeys[strlen(recKeys)-1]=='\r' || recKeys[strlen(recKeys)-1]=='\n'  ){
								recKeys[strlen(recKeys)-1]=0;
							}
							if( recKeys[strlen(recKeys)-1]=='\r' || recKeys[strlen(recKeys)-1]=='\n'  ){
								recKeys[strlen(recKeys)-1]=0;
							}
							syslog( LOG_INFO, "Caught key %s", recKeys );
							sprintf( mybuf, "%s%s", keys, "258EAFA5-E914-47DA-95CA-C5AB0DC85B11" ); 
							memset( keys, 0, sizeof( keys ) );
							SHA1( (unsigned const char *)mybuf, strlen(mybuf), (unsigned char *)keys );
							pp=g_base64_encode( (const guchar *)keys, strlen(keys) );
							memset( keys, 0, sizeof( keys ) );
							strcpy( keys, (const char *)pp );
							g_free( pp );
							if( strcmp( keys, recKeys ) ) {
								syslog( LOG_ERR, "keys not correct %s %s", keys, recKeys );
								status = -1;
							} else {
								syslog( LOG_ERR, "keys correct %s %s", keys, recKeys );
							}
							break;
						}
					}
				}
				if( strlen(mybuf)<=2 ) { // void line at end of headers
					status++;
					syslog( LOG_INFO, "status advanced %d", status );
				}
			}
			if( status == -1 ) {
				syslog( LOG_ERR, "TCP read error" );
				break;
			}
			if( status >= XtNumber(responseHeader)+1 ) { // exit at end of headers
				break;
			}
		}
		if( status == -1 ) {
			close( works->sfd );
			sleep( 2 );
			continue;
		}
		// sending passphrase
		mybuf[0]=0x81;
		mybuf[1]=strlen(works->passphrase);
		mybuf[2]=0x00;
		mybuf[3]=0x00;
		mybuf[4]=0x00;
		mybuf[5]=0x00;
		memcpy( mybuf+6, works->passphrase, strlen( works->passphrase ) );
		n=write( works->sfd, mybuf, strlen( works->passphrase )+6 );
		works->connected = 1;
		works->freqChanged.tv_sec=0;
		works->freqChanged.tv_usec=0;
		works->ritChanged.tv_sec=0;
		works->ritChanged.tv_usec=0;
		works->volumeChanged.tv_sec=0;
		works->volumeChanged.tv_usec=0;
		works->auxVolChanged.tv_sec=0;
		works->auxVolChanged.tv_usec=0;
		works->modeChanged.tv_sec=0;
		works->modeChanged.tv_usec=0;
		works->bwChanged.tv_sec=0;
		works->bwChanged.tv_usec=0;
		works->thresholdChanged.tv_sec=0;
		works->thresholdChanged.tv_usec=0;
		works->agcMlvlChanged.tv_sec=0;
		works->agcMlvlChanged.tv_usec=0;
		works->agcAlvlChanged.tv_sec=0;
		works->agcAlvlChanged.tv_usec=0;
		works->nrChanged.tv_sec=0;
		works->nrChanged.tv_usec=0;
		works->nbChanged.tv_sec=0;
		works->nbChanged.tv_usec=0;
		works->anChanged.tv_sec=0;
		works->anChanged.tv_usec=0;
		works->squelchChanged.tv_sec=0;
		works->squelchChanged.tv_usec=0;
		usleep( 20000 );
		syslog( LOG_INFO, "Websocket Connected" );
		works->setType( works, 0 );	
		works->setDelay( works, 0 );	
		isFirst=-works->buffer_start;
		works->soundWritePointer=0;
		works->soundReadPointer=0;
		works->shutAudio=0;
		memset( works->message, 0, sizeof( works->message) );

		buildConfig( works );
		works->irttmax=0;
		works->irtt=0;
		for( ; works->connected; ) {
			if( works->countTo ) { // if no sound, it breaks and reconnect
				syslog( LOG_INFO, "TCP countTo: sound missing" );
				break;
			}
			memset( mybuf, 0, sizeof( mybuf ) );
			for( n=0; n<4; ) {
				gettimeofday( &start, NULL );
				rc = read( works->sfd, mybuf+n, 4-n );
				myerrno=errno;
				myerrno=myerrno; // only to avoid compilation warnings where syslog is undefined
				gettimeofday( &stop, NULL );
				if( works->countTo ) { // if no sound, it breaks and reconnect
					rc=-2;
					syslog( LOG_INFO, "TCP countTo: sound missing" );
					break;
				}
				if( rc<1 )  {
					syslog( LOG_INFO, "TCP read Error 1A %d %d", rc, myerrno );
					syslog( LOG_INFO, "time= %f mSec\n", (stop.tv_sec-start.tv_sec)*1000+0.001*(stop.tv_usec-start.tv_usec) );
					break;
				}
				n+=rc;
			}
			if( rc<1  )  {
				syslog( LOG_INFO, "TCP read Error 2A %d", errno );
				break;
			}
			// syslog( LOG_INFO, "received %02x %02x %-2.2s %-2.2s", mybuf[0], mybuf[1], mybuf+2, mybuf+4 );
			if( mybuf[1]<0x07e ) {
				pos=(mybuf[1]&0x7F);
				if( (mybuf[0])&0x001 ) {
					pos+=4;
				}
				for( n=0;n<pos-2; ) {
					rc = read( works->sfd, mybuf+4+n, pos-n-2 );
					if( rc<1  )  {
						syslog( LOG_INFO, "TCP read Error 3a %d", errno );
						break;
					}
					n+=rc;
				}
			} else if( mybuf[1]==0x07e ) {
				memcpy( &spos, mybuf+2, 2 );
				spos = ntohs( spos );
				for( n=0;n<spos; ) {
					rc = read( works->sfd, mybuf+n+4, spos-n );
					if( rc<1  )  {
						syslog( LOG_INFO, "TCP read Error 4a %d", errno );
						break;
					}
					n+=rc;
				}
			} else {
				for( n=4; n<6; ) {
					rc = read( works->sfd, mybuf+4, 6-n );
					if( rc<1  )  {
						syslog( LOG_INFO, "TCP read Error 1a %d", errno );
						break;
					}
					n+=rc;
				}
				memcpy( &pos, mybuf+2, 4 );
				pos = ntohl( pos );
				for( n=0;n<pos-2; ) {
					rc = read( works->sfd, mybuf+n+6, pos-n-2 );
					if( rc<1  )  {
						syslog( LOG_INFO, "TCP read Error 5a %d", errno );
						break;
					}
					n+=rc;
				}
			}
			if( rc<1  )  {
				syslog( LOG_INFO, "TCP read Error 6A %d", errno );
				break;
			}
			if( (unsigned char)mybuf[0]==0x82 && (unsigned char)mybuf[1]!=0x7e && !strncmp( mybuf+2, "HB", 2 ) ) {
				gettimeofday( &rec, NULL );
				memcpy( &ns1, mybuf+4, 4 );
				s1=ntohl( ns1 );
				memcpy( &nus1, mybuf+8, 4 );
				us1=ntohl( nus1 );
				memcpy( &ns2, mybuf+12, 4 );
				s2=ntohl( ns2 );
				memcpy( &nus2, mybuf+16, 4 );
				us2=ntohl( nus2 );
				works->ta=1000.0*(s2-s1)+0.001*(us2-us1);
				works->tr=1000.0*(rec.tv_sec-s1)+0.001*(rec.tv_usec-us1);
				works->irtt=1000.0*(rec.tv_sec-s2)+0.001*(rec.tv_usec-us2);
				if( works->irttmax < works->irtt ) {
					works->irttmax=works->irtt;
				}
				works->irttmed=works->irttmed*19/20+works->irtt/20;
			} else if( (unsigned char)mybuf[0]==0x82 && (unsigned char)mybuf[1]!=0x7e && !strncmp( mybuf+2, "CF", 2 ) ) {
				memcpy( &ns1, mybuf+6, 4 );
				s1=ntohl( ns1 );
				memcpy( &ns2, mybuf+10, 4 );
				s2=ntohl( ns2 );
				works->rxcompress=s1;
				s1=works->txcompress;
				works->txcompress=s2;
				rc = opus_encoder_ctl( works->encoder, OPUS_SET_BITRATE( works->txcompress ) );
				if( rc<0 ) {
					syslog( LOG_ERR, "failed to set bitrate: %s", opus_strerror( rc ) );
					works->txcompress=s1;
				}
			} else if( (unsigned char)mybuf[0]==0x82 && (unsigned char)mybuf[1]!=0x7e && !strncmp( mybuf+2, "CA", 2 ) ) {
				memset( works->catbuf, 0, sizeof( catbuf ) );
				memcpy( works->catbuf, mybuf+4, pos-2 );
				works->bytesIn+=pos-2+4;
				decodeCATcmd( works, pos-2, (unsigned char *)works->catbuf );
				if(strncmp( (const char *)works->catbuf, "FP", 2 ) && strncmp( (const char *)works->catbuf, "RP", 2 ) ) {
					if( works->fifoFd > -1 ) {
						rc=write( works->fifoFd, (unsigned char *)works->catbuf, pos-2 );
						if( rc==-1 ) {
							syslog( LOG_ERR, "Not able to write \"%-*.*s\" on %s", pos-2, pos-2, (unsigned char *)works->catbuf, works->setup->FifoDevice );
						}
					}
				}
			} else if( (unsigned char)mybuf[0]==0x82 && (unsigned char)mybuf[1]!=0x7e && !strncmp( mybuf+2, "ST", 2 ) ) {
				memset( works->statbuf, 0, sizeof( catbuf ) );
				memcpy( works->statbuf, mybuf+4, pos-2 );
				works->bytesIn+=pos-2+4;
//for( int j=0; j<44; j++ ) { printf( "%d %d %x %c\n", j, mybuf[4+j], mybuf[4+j], mybuf[4+j] ); }; printf( "\n" ); exit(0);
				storeLinear( works, pos-2, (unsigned char *)mybuf+4 );
			} else if( (unsigned char)mybuf[0]==0x82 && (unsigned char)mybuf[1]==0x7e && !strncmp( mybuf+4, "SP", 2 ) ) {
				memcpy( &pos, mybuf+6, 4 );
				length = ntohl( pos );
				memcpy( &pos, mybuf+10, 4 );
				pos = ntohl( pos );
				memcpy( SPbuf+pos, mybuf+14, pos+512>length?length-pos+512:512 );
				if( pos>=length-512 ) {
					works->bytesIn+=length+6;
					gettimeofday( &(works->lastFFTreceived), NULL );
					storeFFT( works, length, (unsigned char *)SPbuf );
					if( works->plotter->layer != 16 ) {
						gettimeofday( &FFTstop, NULL );
						works->FFTrythm=1000.0*(FFTstop.tv_sec-FFTstart.tv_sec)+0.001*(FFTstop.tv_usec-FFTstart.tv_usec)/(works->FFTdelay+1);
						gettimeofday( &FFTstart, NULL );
					}
				}
			} else if( (unsigned char)mybuf[0]==0x82 && (unsigned char)mybuf[1]!=0x7e && !strncmp( mybuf+2, "PA", 2 ) ) {
				int DataLength=(int)mybuf[1];
				works->bytesIn+=DataLength+4;
				storeParameters( works, DataLength, (unsigned char *)mybuf+4 );
			} else if( (unsigned char)mybuf[0]==0x82 && (unsigned char)mybuf[1]==0x7e && !strncmp( mybuf+4, "FB", 2 ) ) {
				memcpy( &pos, mybuf+6, 4 );
				length = ntohl( pos );
				memcpy( &pos, mybuf+10, 4 );
				pos = ntohl( pos );
				works->bytesIn+=pos+14;
			} else if( (unsigned char)mybuf[0]==0x82 && (unsigned char)mybuf[1]==0x7e && !strncmp( mybuf+4, "SO", 2 ) ) {
				works->countSo++;
				if( works->setup->enableWrite=='1'  && works->sound_buffer ) { // during restat sound_buffer is deleted and reallocated
					memcpy( &pos, mybuf+2, 2 );
					pos = ntohs( pos );
					pos -= 2;
					works->bytesIn+=pos+6;
					if( works->setup->sidetone==0 ) {
						if( works->rxcompress ) {
							pos = opus_decode( works->decoder, (const unsigned char *)mybuf+6, pos, convbuf, works->rx_buffer_frames, 0 );
							if( pos<0 ) {
								memset( works->sound_buffer+works->soundWritePointer*works->rx_buffer_frames*AUDIO_BYTES_PER_SAMPLE, 0, works->rx_buffer_frames*AUDIO_BYTES_PER_SAMPLE );
								pos=works->rx_buffer_frames;
							} else {
								for( j=0; j<2*pos; j++ ) {
									(works->sound_buffer+works->soundWritePointer*pos*4)[2*j]=convbuf[j]&0xFF;
									(works->sound_buffer+works->soundWritePointer*pos*4)[2*j+1]=(convbuf[j]>>8)&0xFF;
								}
							}
						} else {
							memcpy( works->sound_buffer+works->soundWritePointer*works->rx_buffer_frames*AUDIO_BYTES_PER_SAMPLE, mybuf+6, pos );
							pos/=4;
						}
						if( isFirst < 0 ) {
							isFirst++;
						} else {
							works->soundConnected=1;
						}
						works->soundWritePointer++;
						works->soundWritePointer%=(works->buffer_start*2);
					}

				}
			} else if( (unsigned char)mybuf[0]==0x81 && (unsigned char)mybuf[1]==0x7e ) {
				j=mybuf[1]&0x7F;
				if( j==n-6 ) {
					for( n=0; n<j; ) {
						for( k=0; k<4 && n<j; k++ ) {
							mybuf[n+6]^=mybuf[k+2];
							n++;
						}
					}
				}
			} else {
				syslog( LOG_ERR, "wrong message type: %x", mybuf[0] );
				break;
			}
		}
		works->shutAudio=1;
		memset( works->sound_buffer, 0, sizeof( works->sound_buffer ) );
		for( j=0; j<10; j++ ) {
			rc=pthread_mutex_trylock( works->wmut );
			if( rc ) {
				usleep( 1000 );
			} else {
				break;
			}
		}
		if( rc ) {
			syslog( LOG_ERR, "mutex cannot be locked stat frame %d %d", rc, j );
		} else {
			usleep( 10000 );
			works->connected=0;
			syslog( LOG_ERR, "reconnecting" );
			rc=close( works->sfd );
			fclose( bsfd );
			usleep( 10000 );
			rc=pthread_mutex_unlock( works->wmut );
			if( rc ) {
				syslog( LOG_ERR, "mutex cannot be unlocked" );
			}
		}
	}
	works->unterminated=0;
	return NULL;
}

void sendWebSockStatFrame( workp works, char *buffer, int len ) {
	char mybuf[200];
	int rc;
	int j;
	if( !works->connected ) {
		return;
	}
	usleep( 10 );
	for( j=0; j<10; j++ ) {
		rc=pthread_mutex_trylock( works->wmut );
		if( rc ) {
			usleep( 1000 );
		} else {
			break;
		}
	}
	if( rc ) {
		syslog( LOG_ERR, "mutex cannot be locked stat frame %d %d", rc, j );
	} else {
		works->bytesOut+=len+4;
		memcpy( mybuf+4, buffer, len );
		mybuf[0]=0x82;
		mybuf[1]=(char)len+2;
		memcpy( mybuf+2, "ST", 2 );
		if( works->sfd ) {
			write( works->sfd, mybuf, len+4 );
		}
		rc=pthread_mutex_unlock( works->wmut );
		if( rc ) {
			syslog( LOG_ERR, "mutex cannot be unlocked" );
		}
	}
}

void sendWebSockCatFrame( workp works, char *buffer, int len ) {
	char mybuf[200];
	int rc;
	int j;
	if( !works->connected ) {
		return;
	}
	usleep( 10 );
	for( j=0; j<10; j++ ) {
		rc=pthread_mutex_trylock( works->wmut );
		if( rc ) {
			usleep( 1000 );
		} else {
			break;
		}
	}
	if( rc ) {
		syslog( LOG_ERR, "mutex cannot be locked cat frame %d %d", rc, j );
	} else {
		works->bytesOut+=len+4;
		memcpy( mybuf+4, buffer, len );
		mybuf[0]=0x82;
		mybuf[1]=(char)len+2;
		memcpy( mybuf+2, "CA", 2 );
		if( works->sfd ) {
			write( works->sfd, mybuf, len+4 );
		}
		rc=pthread_mutex_unlock( works->wmut );
		if( rc ) {
			syslog( LOG_ERR, "mutex cannot be unlocked" );
		}
	}
}

void sendWebSockSerFrame( workp works, char *buffer, int len ) {
	char mybuf[200];
	int rc;
	int j;
	if( !works->connected ) {
		return;
	}
	usleep( 10 );
	for( j=0; j<10; j++ ) {
		rc=pthread_mutex_trylock( works->wmut );
		if( rc ) {
			usleep( 1000 );
		} else {
			break;
		}
	}
	if( rc ) {
		syslog( LOG_ERR, "mutex cannot be locked serframe %d %d", rc, j );
	} else {
		works->bytesOut+=len+4;
		memcpy( mybuf+4, buffer, len );
		mybuf[0]=0x82;
		mybuf[1]=(char)len+2;
		memcpy( mybuf+2, "SC", 2 );
		if( works->sfd ) {
			write( works->sfd, mybuf, len+4 );
		}
		rc=pthread_mutex_unlock( works->wmut );
		if( rc ) {
			syslog( LOG_ERR, "mutex cannot be unlocked" );
		}
	}
}

void sendWebSockCwFrame( workp works, char *buffer, int len ) {
	char mybuf[200];
	int rc;
	int j;
	if( !works->connected ) {
		return;
	}
	usleep( 10 );
	for( j=0; j<10; j++ ) {
		rc=pthread_mutex_trylock( works->wmut );
		if( rc ) {
			usleep( 1000 );
		} else {
			break;
		}
	}
	if( rc ) {
		syslog( LOG_ERR, "mutex cannot be locked cwframe %d %d", rc, j );
	} else {
		works->bytesOut+=len+4;
		mybuf[0]=0x82;
		mybuf[1]=(char)len+2;
		memcpy( mybuf+2, "CW", 2 );
		memcpy( mybuf+4, buffer, len );
		if( works->sfd ) {
			write( works->sfd, mybuf, len+4 );
		}
		rc=pthread_mutex_unlock( works->wmut );
		if( rc ) {
			syslog( LOG_ERR, "mutex cannot be unlocked" );
		}
	}
}

void sendWebSockCfgFrame( workp works, char *buffer, int len ) {
	char mybuf[200];
	int rc;
	int j;
	if( !works->connected ) {
		return;
	}
	usleep( 10 );
	for( j=0; j<10; j++ ) {
		rc=pthread_mutex_trylock( works->wmut );
		if( rc ) {
			usleep( 1000 );
		} else {
			break;
		}
	}
	if( rc ) {
		syslog( LOG_ERR, "mutex cannot be locked cfgframe %d %d", rc, j );
	} else {
		works->bytesOut+=len+4;
		mybuf[0]=0x82;
		mybuf[1]=(char)len+2;
		memcpy( mybuf+2, "CF", 2 );
		memcpy( mybuf+4, buffer, len );
		if( works->sfd ) {
			write( works->sfd, mybuf, len+4 );
		}
		rc=pthread_mutex_unlock( works->wmut );
		if( rc ) {
			syslog( LOG_ERR, "mutex cannot be unlocked" );
		}
	}
}

void sendWebSockHebFrame( workp works, char *buffer, int len ) {
	char mybuf[200];
	int rc;
	int j;
	int lun;
	struct timeval start;
	long s, us;
	if( !works->connected ) {
		return;
	}
	usleep( 10 );
	for( j=0; j<10; j++ ) {
		rc=pthread_mutex_trylock( works->wmut );
		if( rc ) {
			usleep( 1000 );
		} else {
			break;
		}
	}
	if( rc ) {
		syslog( LOG_ERR, "mutex cannot be locked hebframe %d %d", rc, j );
	} else {
		lun=len;
		len=16;
		works->bytesOut+=len+4;
		mybuf[0]=0x82;
		mybuf[1]=(char)len+2;
		memcpy( mybuf+2, "HB", 2 );
		gettimeofday( &start, NULL );
		s=start.tv_sec;
		us=start.tv_usec;
		s=htonl(s);
		us=htonl(us);
		memcpy( mybuf+4, &s, 4 );
		memcpy( mybuf+8, &us, 4 );
		memset( mybuf+12, 0, 8 );
		if( lun ) {
			memcpy( mybuf+12, buffer, lun );
		}
		if( works->sfd ) {
			write( works->sfd, mybuf, len+4 );
		}
		rc=pthread_mutex_unlock( works->wmut );
		if( rc ) {
			syslog( LOG_ERR, "mutex cannot be unlocked" );
		}
	}
}

void sendWebSockAudioFrame( workp works, char *buffer ) {
	unsigned char mybuf[512000];
	opus_int16 convbuf[51200];
	char intbuf[512000];
	int rc;
	int length;
	short l1;
	int j;
	if( !works->connected ) {
		return;
	}
	if( works->txCwLock == 0 ) {
		if( works->serParms->ptt<1) {
			return;
		}
		if( works->serParms->tune>0 ) {
			return;
		}
	}
	usleep( 10 );
	for( j=0; j<10; j++ ) {
		rc=pthread_mutex_trylock( works->wmut );
		if( rc ) {
			usleep( 1000 );
		} else {
			break;
		}
	}
	if( rc ) {
		syslog( LOG_ERR, "mutex cannot be locked audio frame %d %d", rc, j );
	} else {
		if( works->txcompress ) {
			for( j=0; j<2*works->tx_buffer_frames; j++ ) {
				convbuf[j]=(unsigned char)(buffer[2*j+1])<<8|(unsigned char)buffer[2*j];
			}
			length = opus_encode( works->encoder, convbuf, works->tx_buffer_frames, mybuf, works->tx_buffer_frames * AUDIO_BYTES_PER_SAMPLE );
			if( length<0 ) {
				syslog(  LOG_INFO, "encode failed: %s\n", opus_strerror( length ) );
			} else {
				memcpy( intbuf, mybuf, length );
			}
		} else {
			length = works->tx_buffer_frames * AUDIO_BYTES_PER_SAMPLE;
			memcpy( intbuf, buffer, length );
		}
		if( length > 0 ) {
			works->bytesOut+=length+6;
			l1=(short)length+2;
			l1=htons(l1);
			mybuf[0]=0x82;
			mybuf[1]=0x7e;
			memcpy( mybuf+2, &l1, 2 );
			memcpy( mybuf+4, "SO", 2 );
			memcpy( mybuf+6, intbuf, length );
			if( works->sfd ) {
				rc=write( works->sfd, mybuf, length+6 );
				if( rc != length+6 ) {
					syslog( LOG_ERR, "rc(%d) != length+6(%d)", rc, length+6);
				}
			}
		}
		rc=pthread_mutex_unlock( works->wmut );
		if( rc ) {
			syslog( LOG_ERR, "mutex cannot be unlocked" );
		}
	}
}

void buildConfig( workp works ) {
	char buffer[30];
	long rxh, txh, rxn, txn;
	rxh=works->rxcompress;
	txh=works->txcompress;
	rxn=htonl(rxh);
	txn=htonl(txh);
	memcpy( buffer, "1", 1 );	// opus encoder
	memcpy( buffer+1, "1", 1 );	// opus decoder
	memcpy( buffer+2, &rxn, 4 );
	memcpy( buffer+6, &txn, 4 );
	rxh=1000*works->setup->rxChunkTime;
	works->rx_buffer_frames=works->setup->rxChunkTime*SAMPLES/2.5;
	txh=1000*works->setup->txChunkTime;
	works->tx_buffer_frames=works->setup->txChunkTime*SAMPLES/2.5;
	rxn=htonl(rxh);
	txn=htonl(txh);
	memcpy( buffer+10, &rxn, 4 );
	memcpy( buffer+14, &txn, 4 );
	rxh=(long)(works->setup->bufferingTime/works->setup->txChunkTime);
	if( rxh<1 ) {
		rxh=1;
	}
	rxh=htonl(rxh);
	memcpy( buffer+18, &rxh, 4 );
	sendWebSockCfgFrame( works, buffer, 22 );
}
